# Copyright (c) Acconeer AB, 2022-2025
# All rights reserved

from __future__ import annotations

import enum
import warnings
from typing import Any, Dict, List, Optional, Tuple

import attrs
import h5py
import numpy as np
import numpy.typing as npt
from attributes_doc import attributes_doc
from packaging.version import Version

from acconeer.exptool import a121, opser
from acconeer.exptool import type_migration as tm
from acconeer.exptool._core.class_creation.attrs import attrs_optional_ndarray_isclose
from acconeer.exptool.a121._core import utils
from acconeer.exptool.a121._core.entities.configs.sensor_config import (
    VALIDATION_TAG_BUFFER_SIZE_TOO_LARGE,
)
from acconeer.exptool.a121._h5_utils import _create_h5_string_dataset
from acconeer.exptool.a121.algo import (
    AlgoConfigBase,
    Controller,
    PeakSortingMethod,
    ReflectorShape,
)

from ._aggregator import Aggregator, AggregatorConfig, ProcessorSpec
from ._context import (
    CloseRangeCalibration,
    DetectorContext,
    NoiseCalibration,
    OffsetCalibration,
    RecordedThresholdCalibration,
    detector_context_timeline,
)
from ._processors import (
    DEFAULT_FIXED_AMPLITUDE_THRESHOLD_VALUE,
    DEFAULT_FIXED_STRENGTH_THRESHOLD_VALUE,
    DEFAULT_THRESHOLD_SENSITIVITY,
    MeasurementType,
    ProcessorContext,
    ProcessorResult,
    ThresholdMethod,
)
from ._translation import (
    detector_config_to_processor_specs,
    detector_config_to_session_config,
    get_num_far_subsweeps,
)


class ConfigMismatchError(RuntimeError):
    pass


@attrs.frozen(kw_only=True)
class DetectorStatus:
    detector_state: DetailedStatus
    ready_to_start: bool


class DetailedStatus(enum.Enum):
    OK = enum.auto()
    END_LESSER_THAN_START = enum.auto()
    SENSOR_IDS_NOT_UNIQUE = enum.auto()
    CONTEXT_MISSING = enum.auto()
    CALIBRATION_MISSING = enum.auto()
    CONFIG_MISMATCH = enum.auto()


PRF_REMOVED_ET_VERSION = Version("v7.13.2")


@attributes_doc
@attrs.mutable(kw_only=True)
class DetectorConfig(AlgoConfigBase):
    start_m: float = attrs.field(default=0.25)
    """Start of measurement range in meters."""

    end_m: float = attrs.field(default=3.0)
    """End of measurement range in meters."""

    max_step_length: Optional[int] = attrs.field(default=None)
    """If set, limits the step length.

    If no argument is provided, the step length is automatically calculated based on the profile.

    Reducing the step length increases SNR through more efficient distance filtering, while
    increasing the measurement time and the processing load.
    """

    max_profile: a121.Profile = attrs.field(default=a121.Profile.PROFILE_5, converter=a121.Profile)
    """Specifies the longest allowed profile.

    If no argument is provided, the highest possible profile without interference of direct
    leakage is used to maximize SNR.

    A lower profile improves the radial resolution.
    """

    close_range_leakage_cancellation: bool = attrs.field(default=False)
    """Enable close range leakage cancellation logic.

    Close range leakage cancellation refers to the process of measuring close to the
    sensor(<100mm) by first characterizing the direct leakage, and then subtracting it
    from the measured sweep in order to isolate the signal component of interest.

    The close range leakage cancellation process requires the sensor to be installed in its
    intended geometry with free space in front of the sensor during detector calibration.
    """

    signal_quality: float = attrs.field(default=15.0)
    """Signal quality (dB).

    High quality equals higher HWAAS and better SNR but increases power consumption."""

    threshold_method: ThresholdMethod = attrs.field(
        default=ThresholdMethod.CFAR,
        converter=ThresholdMethod,
    )
    """Threshold method"""

    peaksorting_method: PeakSortingMethod = attrs.field(
        default=PeakSortingMethod.STRONGEST,
        converter=PeakSortingMethod,
    )
    """Sorting method of estimated distances.

    The distance estimates are sorted according to the selected strategy, before being return
    by th application.
    """

    reflector_shape: ReflectorShape = attrs.field(
        default=ReflectorShape.GENERIC,
        converter=ReflectorShape,
    )
    """Reflector shape."""

    num_frames_in_recorded_threshold: int = attrs.field(default=100)
    """Number of frames used when calibrating threshold.

    A lower number reduce calibration time and a higher number results in a more statistically
    significant threshold.
    """

    fixed_threshold_value: float = attrs.field(default=DEFAULT_FIXED_AMPLITUDE_THRESHOLD_VALUE)
    """Value of fixed amplitude threshold."""

    fixed_strength_threshold_value: float = attrs.field(
        default=DEFAULT_FIXED_STRENGTH_THRESHOLD_VALUE
    )
    """Value of fixed strength threshold."""

    threshold_sensitivity: float = attrs.field(default=DEFAULT_THRESHOLD_SENSITIVITY)
    """Sensitivity of threshold.

    High sensitivity equals low detection threshold, low sensitivity equals high detection
    threshold."""

    update_rate: Optional[float] = attrs.field(default=50.0)
    """Sets the detector update rate."""

    @start_m.validator
    def _(self, _: Any, value: float) -> None:
        if value < Detector.MIN_DIST_M:
            msg = f"Cannot start measurements closer than {Detector.MIN_DIST_M}m"
            raise ValueError(msg)

    @end_m.validator
    def _(self, _: Any, value: float) -> None:
        if value > Detector.MAX_DIST_M:
            msg = f"Cannot measure further than {Detector.MAX_DIST_M}m"
            raise ValueError(msg)

    def _collect_validation_results(self) -> list[a121.ValidationResult]:
        validation_results: list[a121.ValidationResult] = []

        if self.end_m < self.start_m:
            validation_results.append(
                a121.ValidationError(
                    self,
                    "start_m",
                    "Must be smaller than 'Range end'",
                )
            )

            validation_results.append(
                a121.ValidationError(
                    self,
                    "end_m",
                    "Must be greater than 'Range start'",
                )
            )

        if self.max_step_length is not None and (
            not utils.is_divisor_of(24, self.max_step_length)
            and not utils.is_multiple_of(24, self.max_step_length)
        ):
            valid_step_length = next(
                sl
                for sl in range(self.max_step_length, 0, -1)
                if utils.is_divisor_of(24, sl) or utils.is_multiple_of(24, sl)
            )
            validation_results.append(
                a121.ValidationWarning(
                    self,
                    "max_step_length",
                    "Actual step length will be rounded down "
                    + f"to the closest valid step length ({valid_step_length}).",
                )
            )

        session_config = detector_config_to_session_config(self, [1])
        if any(
            sc_val_res.tag is VALIDATION_TAG_BUFFER_SIZE_TOO_LARGE
            for sc_val_res in session_config._collect_validation_results()
        ):
            error_msg_stem = "Required buffer size is too large. "
            validation_results.append(
                a121.ValidationError(
                    self,
                    "end_m",
                    error_msg_stem + "Try decreasing the range.",
                )
            )
            validation_results.append(
                a121.ValidationError(
                    self,
                    "start_m",
                    error_msg_stem + "Try decreasing the range.",
                )
            )
            validation_results.append(
                a121.ValidationError(
                    self,
                    "max_step_length",
                    (
                        error_msg_stem
                        + "Allowing a higher step length will require a smaller buffer."
                    ),
                )
            )

        return validation_results


@attrs.frozen(kw_only=True)
class DetectorResult:
    distances: Optional[npt.NDArray[np.float64]] = attrs.field(
        default=None, eq=attrs_optional_ndarray_isclose
    )
    """Estimated distances (m), sorted according to the selected peak sorting strategy."""

    strengths: Optional[npt.NDArray[np.float64]] = attrs.field(
        default=None, eq=attrs_optional_ndarray_isclose
    )
    """Estimated reflector strengths (dB) corresponding to the peak amplitude of the estimated
    distances.
    """

    near_edge_status: Optional[bool] = attrs.field(default=None)
    """Boolean indicating an object close to the start edge, located outside of the
    measurement range.
    """

    calibration_needed: Optional[bool] = attrs.field(default=None)
    """Indication of calibration needed. The sensor calibration needs to be redone if this
    indication is set.

    A sensor calibration should be followed by a detector calibration update, by calling
    :func:`update_detector_calibration`.
    """

    temperature: Optional[int] = attrs.field(default=None)
    """Temperature in sensor during measurement (in degree Celsius). Notice that this has poor
    absolute accuracy.
    """

    processor_results: List[ProcessorResult] = attrs.field()
    """Processing result. Used for visualization in Exploration Tool."""

    service_extended_result: List[Dict[int, a121.Result]] = attrs.field()
    """Service extended result. Used for visualization in Exploration Tool."""


class Detector(Controller[DetectorConfig, Dict[int, DetectorResult]]):
    """Distance detector
    :param client: Client
    :param sensor_id: Sensor id
    :param detector_config: Detector configuration
    :param context: Detector context
    """

    MIN_DIST_M = 0.0
    MAX_DIST_M = 23.0

    session_config: a121.SessionConfig
    processor_specs: List[ProcessorSpec]
    context: DetectorContext

    def __init__(
        self,
        *,
        client: a121.Client,
        sensor_ids: list[int],
        detector_config: DetectorConfig,
        context: Optional[DetectorContext] = None,
    ) -> None:
        super().__init__(client=client, config=detector_config)
        self.sensor_ids = sensor_ids
        self.started = False

        if context is None or not bool(context.sensor_ids):
            self.context = DetectorContext(sensor_ids=self.sensor_ids)
        else:
            self.context = context

        self.aggregator: Optional[Aggregator] = None

        self.update_config(self.config)

    def _validate_ready_for_calibration(self) -> None:
        if self.started:
            msg = "Already started"
            raise RuntimeError(msg)
        if self.processor_specs is None:
            msg = "Processor specification not defined"
            raise ValueError(msg)
        if self.session_config is None:
            msg = "Session config not defined"
            raise ValueError(msg)

    def calibrate_detector(self) -> None:
        """Run the required detector calibration routines, based on the detector config."""

        self._validate_ready_for_calibration()

        self.context.offset_calibration = OffsetCalibration.create(self.client, self.sensor_ids)

        self.context.noise_calibration = NoiseCalibration.create(
            self.client, self.sensor_ids, self.session_config
        )

        if self._has_close_range_measurement(self.config):
            self.context.close_range_calibration = CloseRangeCalibration.create(
                self.client, self.session_config
            )

        if self._has_close_range_measurement(self.config) or self._has_recorded_threshold_mode(
            self.config, self.sensor_ids
        ):
            self.context.recorded_threshold_calibration = RecordedThresholdCalibration.create(
                self.client, self.session_config, self.config.num_frames_in_recorded_threshold
            )

        self.context.session_config_used_during_calibration = self.session_config

    def update_detector_calibration(self) -> None:
        """Do a detector calibration update by running a subset of the calibration routines.

        Once the detector is calibrated, by calling :func:`calibrate_detector`, a sensor
        calibration should be followed by a detector calibration update.
        """

        self._validate_ready_for_calibration()

        self.context.offset_calibration = OffsetCalibration.create(self.client, self.sensor_ids)

    @staticmethod
    def _get_sensor_calibrations(context: DetectorContext) -> dict[int, a121.SensorCalibration]:
        return (
            context.close_range_calibration.sensor_calibrations
            if context.close_range_calibration is not None
            else {}
        )

    @classmethod
    def get_detector_status(
        cls,
        config: DetectorConfig,
        context: DetectorContext,
        sensor_ids: list[int],
    ) -> DetectorStatus:
        """Returns the detector status along with the detector state."""

        if config.end_m < config.start_m:
            return DetectorStatus(
                detector_state=DetailedStatus.END_LESSER_THAN_START,
                ready_to_start=False,
            )

        if len(sensor_ids) != len(set(sensor_ids)):
            return DetectorStatus(
                detector_state=DetailedStatus.SENSOR_IDS_NOT_UNIQUE,
                ready_to_start=False,
            )

        if len(context.sensor_ids) == 0:
            return DetectorStatus(
                detector_state=DetailedStatus.CONTEXT_MISSING,
                ready_to_start=False,
            )

        session_config = detector_config_to_session_config(config=config, sensor_ids=sensor_ids)

        # Offset calibration is always performed as a part of the detector calibration process.
        # Use this as indication whether detector calibration has been performed.
        calibration_missing = np.any(
            [context.offset_calibration is None for sensor_id in context.sensor_ids]
        )
        config_mismatch = context.session_config_used_during_calibration != session_config

        if calibration_missing:
            detector_state = DetailedStatus.CALIBRATION_MISSING
        elif config_mismatch:
            detector_state = DetailedStatus.CONFIG_MISMATCH
        else:
            detector_state = DetailedStatus.OK

        return DetectorStatus(
            detector_state=detector_state,
            ready_to_start=(detector_state == DetailedStatus.OK),
        )

    @staticmethod
    def _close_range_calibrated(context: DetectorContext) -> bool:
        return context.close_range_calibration is not None

    @staticmethod
    def _recorded_threshold_calibrated(context: DetectorContext) -> bool:
        return context.recorded_threshold_calibration is not None

    @classmethod
    def _has_close_range_measurement(cls, config: DetectorConfig) -> bool:
        # sensor_ids=[1] as the detector is running the same config for all sensors.
        session_config = detector_config_to_session_config(config, sensor_ids=[1])
        num_far_subsweeps = get_num_far_subsweeps(session_config, config)
        specs = detector_config_to_processor_specs(
            config=config, sensor_ids=[1], num_far_subsweeps=num_far_subsweeps
        )
        return MeasurementType.CLOSE_RANGE in [
            spec.processor_config.measurement_type for spec in specs
        ]

    @classmethod
    def _has_recorded_threshold_mode(cls, config: DetectorConfig, sensor_ids: list[int]) -> bool:
        processor_specs = detector_config_to_processor_specs(
            config=config,
            sensor_ids=sensor_ids,
            # num_far_subsweeps hardcoded as it does not impact the threshold method.
            num_far_subsweeps=4,
        )
        return ThresholdMethod.RECORDED in [
            spec.processor_config.threshold_method for spec in processor_specs
        ]

    def start(
        self,
        recorder: Optional[a121.Recorder] = None,
        *,
        _algo_group: Optional[h5py.Group] = None,
    ) -> None:
        """Method for setting up measurement session."""

        if self.started:
            msg = "Already started"
            raise RuntimeError(msg)

        status = self.get_detector_status(self.config, self.context, self.sensor_ids)

        if not status.ready_to_start:
            msg = f"Not ready to start ({status.detector_state.name})"
            if status.detector_state.name == DetailedStatus.CONFIG_MISMATCH.name:
                raise ConfigMismatchError(msg)
            else:
                raise RuntimeError(msg)
        specs = self._add_context_to_processor_spec()

        sensor_calibration = self._get_sensor_calibrations(self.context)

        extended_metadata = self.client.setup_session(
            self.session_config, calibrations=sensor_calibration
        )

        assert isinstance(extended_metadata, list)
        assert np.all(
            [self.context.offset_calibration is not None for sensor_id in self.context.sensor_ids]
        )
        aggregator_config = AggregatorConfig(peak_sorting_method=self.config.peaksorting_method)
        self.aggregators = {
            sensor_id: Aggregator(
                session_config=self.session_config,
                extended_metadata=extended_metadata,
                config=aggregator_config,
                specs=specs[sensor_id],
                sensor_id=sensor_id,
            )
            for sensor_id in self.sensor_ids
        }

        if recorder is not None:
            if isinstance(recorder, a121.H5Recorder):
                if _algo_group is None:
                    _algo_group = recorder.require_algo_group("distance_detector")

                _record_algo_data(
                    _algo_group,
                    self.sensor_ids,
                    self.config,
                    self.context,
                )
            else:
                # Should never happen as we currently only have the H5Recorder
                warnings.warn("Will not save algo data")

            self.client.attach_recorder(recorder)

        self.client.start_session()
        self.started = True

    def get_next(self) -> Dict[int, DetectorResult]:
        """Called from host to get next measurement."""
        if not self.started:
            msg = "Not started"
            raise RuntimeError(msg)

        assert self.aggregators is not None

        extended_result = self.client.get_next()
        assert isinstance(extended_result, list)

        aggregator_results = {
            sensor_id: self.aggregators[sensor_id].process(extended_result=extended_result)
            for sensor_id in self.sensor_ids
        }

        result = {
            sensor_id: DetectorResult(
                strengths=aggregator_results[sensor_id].estimated_strengths,
                distances=aggregator_results[sensor_id].estimated_distances,
                near_edge_status=aggregator_results[sensor_id].near_edge_status,
                calibration_needed=extended_result[0][sensor_id].calibration_needed,
                temperature=extended_result[0][sensor_id].temperature,
                processor_results=aggregator_results[sensor_id].processor_results,
                service_extended_result=aggregator_results[sensor_id].service_extended_result,
            )
            for sensor_id in self.sensor_ids
        }

        return result

    def update_config(self, config: DetectorConfig) -> None:
        """Updates the session config and processor specification based on the detector
        configuration."""
        self.session_config = detector_config_to_session_config(config, self.sensor_ids)
        num_far_subsweeps = get_num_far_subsweeps(self.session_config, config)
        self.processor_specs = detector_config_to_processor_specs(
            config, self.sensor_ids, num_far_subsweeps
        )

    def stop(self) -> Any:
        """Stops the measurement session."""
        if not self.started:
            msg = "Already stopped"
            raise RuntimeError(msg)

        self.client.stop_session()
        recorder = self.client.detach_recorder()
        if recorder is None:
            recorder_result = None
        else:
            recorder_result = recorder.close()

        self.started = False

        return recorder_result

    def stop_recorder(self) -> Any:
        """Stop the recorder if the detector session is already stopped"""
        recorder = self.client.detach_recorder()
        if recorder is None:
            recorder_result = None
        else:
            recorder_result = recorder.close()

        return recorder_result

    def stop_detector(self) -> Any:
        """Stops only the detector session."""
        if not self.started:
            msg = "Already stopped"
            raise RuntimeError(msg)

        self.client.stop_session()
        self.started = False
        return None

    def _add_context_to_processor_spec(self) -> dict[int, list[ProcessorSpec]]:
        """
        Create and add processor context to processor specification.
        """

        updated_specs_all_sensors = {}
        for sensor_id in self.context.sensor_ids:
            assert self.context.noise_calibration is not None
            bg_noise_stds = self.context.noise_calibration.bg_noise_std(
                sensor_id, self.processor_specs, self.session_config
            )
            updated_specs: List[ProcessorSpec] = []

            if self.context.recorded_threshold_calibration is not None:
                recorded_thresholds_mean_sweep = (
                    self.context.recorded_threshold_calibration.recorded_thresholds_mean_sweep(
                        sensor_id,
                        self.processor_specs,
                        self.session_config,
                        self.context.noise_calibration,
                        self.context.close_range_calibration,
                    )
                )
                recorded_thresholds_noise_std = (
                    self.context.recorded_threshold_calibration.recorded_thresholds_noise_std(
                        sensor_id,
                        self.processor_specs,
                        self.session_config,
                        self.context.noise_calibration,
                        self.context.close_range_calibration,
                    )
                )
            else:
                recorded_thresholds_mean_sweep = None
                recorded_thresholds_noise_std = None

            for idx, (spec, bg_noise_std) in enumerate(zip(self.processor_specs, bg_noise_stds)):
                if (
                    recorded_thresholds_mean_sweep is not None
                    and recorded_thresholds_noise_std is not None
                ):
                    recorded_threshold_mean_sweep = recorded_thresholds_mean_sweep[idx]
                    recorded_threshold_noise_std = recorded_thresholds_noise_std[idx]
                else:
                    recorded_threshold_mean_sweep = None
                    recorded_threshold_noise_std = None

                if self.context.close_range_calibration is not None:
                    direct_leakage = self.context.close_range_calibration.direct_leakage(
                        sensor_id,
                        self.processor_specs,
                        self.session_config,
                    )
                    phase_jitter_comp_ref = (
                        self.context.close_range_calibration.phase_jitter_comp_reference(
                            sensor_id,
                            self.processor_specs,
                            self.session_config,
                        )
                    )
                else:
                    direct_leakage = None
                    phase_jitter_comp_ref = None

                if self.context.offset_calibration is not None:
                    loopback_peak_location_m = (
                        self.context.offset_calibration.loopback_peak_location_m(sensor_id)
                    )
                else:
                    loopback_peak_location_m = None

                if self.context.recorded_threshold_calibration is not None:
                    reference_temperature = (
                        self.context.recorded_threshold_calibration.reference_temperature(
                            sensor_id,
                            self.processor_specs,
                            self.session_config,
                            self.context.noise_calibration,
                            self.context.close_range_calibration,
                        )
                    )
                else:
                    reference_temperature = None

                processor_context = ProcessorContext(
                    recorded_threshold_mean_sweep=recorded_threshold_mean_sweep,
                    recorded_threshold_noise_std=recorded_threshold_noise_std,
                    bg_noise_std=bg_noise_std,
                    direct_leakage=direct_leakage,
                    phase_jitter_comp_ref=phase_jitter_comp_ref,
                    reference_temperature=reference_temperature,
                    loopback_peak_location_m=loopback_peak_location_m,
                )
                updated_specs.append(attrs.evolve(spec, processor_context=processor_context))
            updated_specs_all_sensors[sensor_id] = updated_specs
        return updated_specs_all_sensors


def _record_algo_data(
    algo_group: h5py.Group,
    sensor_ids: list[int],
    config: DetectorConfig,
    context: DetectorContext,
) -> None:
    algo_group.create_dataset(
        "sensor_ids",
        data=sensor_ids,
        track_times=False,
    )
    _create_h5_string_dataset(algo_group, "detector_config", config.to_json())

    context_group = algo_group.create_group("context")
    opser.serialize(context, context_group)


def _load_algo_data(
    algo_group: h5py.Group,
) -> Tuple[list[int], DetectorConfig, DetectorContext]:
    sensor_ids = algo_group["sensor_ids"][()].tolist()
    try:
        config = detector_config_timeline.migrate(algo_group["detector_config"][()].decode())
    except tm.core.MigrationErrorGroup as exc:
        msg = ""
        match = exc.subgroup(BadMigrationPathError)
        if match is not None:
            # Add more details from exception thrown
            for exc_arg in match.exceptions[0].args:
                if isinstance(exc_arg, str):
                    print(exc_arg)
                    msg += f", {exc_arg}"

        raise TypeError(msg) from exc

    context_group = algo_group["context"]
    context = detector_context_timeline.migrate(context_group)

    return sensor_ids, config, context


@attrs.mutable
@attrs.mutable(kw_only=True)
class _DetectorConfig_v0(AlgoConfigBase):
    start_m: float = attrs.field(default=0.25)
    end_m: float = attrs.field(default=3.0)
    max_step_length: Optional[int] = attrs.field(default=None)
    max_profile: a121.Profile = attrs.field(default=a121.Profile.PROFILE_5, converter=a121.Profile)
    close_range_leakage_cancellation: bool = attrs.field(default=False)
    signal_quality: float = attrs.field(default=15.0)
    threshold_method: ThresholdMethod = attrs.field(
        default=ThresholdMethod.CFAR,
        converter=ThresholdMethod,
    )
    peaksorting_method: PeakSortingMethod = attrs.field(
        default=PeakSortingMethod.STRONGEST,
        converter=PeakSortingMethod,
    )
    reflector_shape: ReflectorShape = attrs.field(
        default=ReflectorShape.GENERIC,
        converter=ReflectorShape,
    )
    num_frames_in_recorded_threshold: int = attrs.field(default=100)
    fixed_threshold_value: float = attrs.field(default=DEFAULT_FIXED_AMPLITUDE_THRESHOLD_VALUE)
    fixed_strength_threshold_value: float = attrs.field(
        default=DEFAULT_FIXED_STRENGTH_THRESHOLD_VALUE
    )
    threshold_sensitivity: float = attrs.field(default=DEFAULT_THRESHOLD_SENSITIVITY)
    update_rate: Optional[float] = attrs.field(default=50.0)
    prf: Optional[a121.PRF] = attrs.field(default=None)

    def _collect_validation_results(self) -> list[a121.ValidationResult]:
        return []


class BadMigrationPathError(Exception): ...


def always_raise_an_error_when_migrating_v0(_: _DetectorConfig_v0) -> DetectorConfig:
    msg = f"Try opening the file in an earlier version of ET <= {PRF_REMOVED_ET_VERSION}"
    raise BadMigrationPathError(msg)


detector_config_timeline = (
    tm.start(_DetectorConfig_v0)
    .load(str, _DetectorConfig_v0.from_json, fail=[TypeError])
    .nop()
    .epoch(DetectorConfig, always_raise_an_error_when_migrating_v0, fail=[BadMigrationPathError])
    .load(str, DetectorConfig.from_json, fail=[TypeError])
)



# Copyright (c) Acconeer AB, 2022-2025
# All rights reserved

from __future__ import annotations

import typing as t

from PySide6.QtWidgets import QVBoxLayout, QWidget

from acconeer.exptool.app.new.app_model import AppModel
from acconeer.exptool.app.new.ui import utils

from .ui_plugin_base import UiPluginBase


class ViewPluginBase(UiPluginBase):
    def __init__(self, app_model: AppModel) -> None:
        super().__init__(app_model)

        self.app_model = app_model

        self.app_model.sig_notify.connect(self.on_app_model_update)
        self.app_model.sig_backend_state_changed.connect(self.on_backend_state_update)
        self.app_model.sig_load_plugin.connect(self.on_load_plugin)

        layout = QVBoxLayout()

        self._sticky_widget = QWidget()
        self._scrolly_widget = QWidget()

        layout.addWidget(self._sticky_widget)
        layout.addWidget(utils.HorizontalSeparator())
        layout.addWidget(
            utils.ScrollAreaDecorator(
                utils.TopAlignDecorator(
                    self._scrolly_widget,
                )
            )
        )

        self.setLayout(layout)

    def on_load_plugin(self, plugin_spec: t.Optional[t.Any]) -> None:
        if plugin_spec is None:
            self.app_model.sig_notify.disconnect(self.on_app_model_update)
            self.app_model.sig_backend_state_changed.disconnect(self.on_backend_state_update)
            self.app_model.sig_load_plugin.disconnect(self.on_load_plugin)

    def on_app_model_update(self, app_model: AppModel) -> None:
        pass

    def on_backend_state_update(self, state: t.Optional[t.Any]) -> None:
        pass

    @property
    def sticky_widget(self) -> QWidget:
        """The sticky widget. The sticky area is located at the top."""
        return self._sticky_widget

    @property
    def scrolly_widget(self) -> QWidget:
        """The scrolly widget. The scrolly area is located below the sticky area"""
        return self._scrolly_widget



# Copyright (c) Acconeer AB, 2023-2024
# All rights reserved

from __future__ import annotations

from typing import Optional, Tuple


def unwrap_ticks(
    ticks: list[int], minimum_tick: Optional[int], limit: int = 2**32
) -> Tuple[list[int], Optional[int]]:
    """Unwraps a sequence of ticks belonging to a collection of result

    The server tick (attached to produced Results) wraps at 2^32 (uint32_t). Thus, if the raw tick is
    used for evaluating the time between two results, it will be incorrect if a wrap has occurred
    between them. Therefore, it has to be accounted for by "unwrapping".

    Wrapping can occur between the results in an extended result, and that the results are not
    necessarily ordered by the tick. This means that we have to look at all the ticks produced in
    the extended result at the same time.

    For example:

    Let's say that the most recent tick we saw was 70 and the wrap happens at limit = 100. Given 2 results
    with ticks 10 and 90; it's more likely that 10 has wrapped than not, we assume it's
    actually after 90, i.e., 110:

    >>> unwrap_ticks([10, 90], minimum_tick=70, limit=100)
    ([110, 90], ...)

    Now, let's also consider that the previous maximum unwrapped tick was 195, which is now the
    'minimum tick'. From this, we know that the ticks have wrapped before and can account for that,
    resulting in the final unwrapped ticks of 310 and 290. The new minimum tick will then be 310:

    >>> unwrap_ticks([10, 90], minimum_tick=195, limit=100)
    ([310, 290], 310)
    """

    if len(ticks) == 0:
        return [], None

    if any(tick < 0 or tick >= limit for tick in ticks):
        msg = "Tick value out of bounds"
        raise ValueError(msg)

    if (max(ticks) - min(ticks)) > limit // 2:
        ticks = [tick + limit if tick < limit // 2 else tick for tick in ticks]

    if minimum_tick is not None:
        num_wraps = max((minimum_tick - tick - 1) // limit + 1 for tick in ticks)
        ticks = [num_wraps * limit + tick for tick in ticks]

    return ticks, max(ticks)



# Copyright (c) Acconeer AB, 2023-2026
# All rights reserved
from __future__ import annotations

import contextlib
import time
from typing import Any, Optional

import libusb_package
import usb.core
from usb.util import CTRL_RECIPIENT_INTERFACE, CTRL_TYPE_CLASS

from .buffered_link import BufferedLink, LinkError


def get_libusb_backend() -> Any:
    """Helper function to setup libusb backend"""
    return libusb_package.get_libusb1_backend()


class UsbPortError(Exception):
    pass


class PyUsbCdc:
    USB_CDC_CMD_SEND_BREAK = 0x23
    USB_MESSAGE_TIMEOUT = 2
    USB_MESSAGE_TIMEOUT_MS = 1000 * USB_MESSAGE_TIMEOUT
    USB_PACKET_TIMEOUT_MS = 200

    def __init__(
        self,
        vid: Optional[int] = None,
        pid: Optional[int] = None,
        serial: Optional[str] = None,
        start: bool = True,
    ) -> None:
        self.serial = serial
        self.vid = vid
        self.pid = pid
        self._dev = None
        self._iface = None
        self._cdc_data_out_ep = None
        self._cdc_data_in_ep = None
        self._rxremaining = b""
        self._backend = get_libusb_backend()

        if not (vid and pid):
            msg = "Must provide vid & pid of device to connect to"
            raise AttributeError(msg)

        if start:
            self.open()

    def open(self) -> bool:
        def match(dev: usb.core.Device) -> bool:
            if self.vid == dev.idVendor and self.pid == dev.idProduct:
                if self.serial is not None:
                    with contextlib.suppress(ValueError):
                        serial_number = dev.serial_number
                        if serial_number and serial_number == self.serial:
                            return True
                else:
                    return True
            return False

        self._dev = usb.core.find(custom_match=match, backend=self._backend)

        if self._dev is None:
            msg = "Port is not open"
            raise UsbPortError(msg)
        try:
            self._dev.reset()
            self._dev.set_configuration()
        except usb.core.USBError:
            msg = "Could not access USB device, are USB permissions setup correctly?"
            raise UsbPortError(msg)
        except NotImplementedError:
            # The function set_configuration is not implemented in windows libusb
            pass

        config = self._dev.get_active_configuration()

        # Interface 1 is the CDC Data interface
        self._iface = config.interfaces()[1]
        try:
            usb.util.claim_interface(self._dev, self._iface.bInterfaceNumber)
        except usb.core.USBError:
            msg = "Could not claim USB device interface, is somebody else using it?"
            raise UsbPortError(msg)
        # Interface 0 has one Endpoint for data out and one Endpoint for data in
        self._cdc_data_out_ep = self._iface.endpoints()[0]
        self._cdc_data_in_ep = self._iface.endpoints()[1]

        return True

    def read(self, size: Optional[int] = None) -> list[int]:
        if self._dev is None:
            msg = "Port is not open"
            raise UsbPortError(msg)

        rx = [self._rxremaining]
        length = len(self._rxremaining)
        self._rxremaining = b""
        end_timeout = time.monotonic() + self.USB_MESSAGE_TIMEOUT
        if size is not None:
            while length < size:
                with contextlib.suppress(usb.core.USBTimeoutError):
                    c = self._dev.read(
                        self._cdc_data_in_ep.bEndpointAddress,
                        self._cdc_data_in_ep.wMaxPacketSize,
                        timeout=self.USB_PACKET_TIMEOUT_MS,
                    )
                if c is not None and len(c):
                    rx.append(c)
                    length += len(c)
                if time.monotonic() > end_timeout:
                    break
        else:
            c = None
            with contextlib.suppress(usb.core.USBTimeoutError):
                c = self._dev.read(
                    self._cdc_data_in_ep.bEndpointAddress,
                    self._cdc_data_in_ep.wMaxPacketSize,
                    timeout=self.USB_MESSAGE_TIMEOUT_MS,
                )
            if c is not None and len(c):
                rx.append(c)
        chunk = b"".join(rx)
        if size and len(chunk) >= size:
            if self._rxremaining:
                self._rxremaining = chunk[size:] + self._rxremaining
            else:
                self._rxremaining = chunk[size:]
            chunk = chunk[0:size]
        return chunk

    def write(self, data: bytes) -> None:
        if self._dev is None:
            msg = "Port is not open"
            raise UsbPortError(msg)
        self._dev.write(self._cdc_data_out_ep, data)

    def reset_input_buffer(self) -> None:
        if self._dev is None:
            msg = "Port is not open"
            raise UsbPortError(msg)

        self._rxremaining = b""
        while True:
            try:
                self._dev.read(
                    self._cdc_data_in_ep.bEndpointAddress,
                    self._cdc_data_in_ep.wMaxPacketSize,
                    timeout=self.USB_PACKET_TIMEOUT_MS,
                )
            except usb.core.USBTimeoutError:
                break

    def send_break(self, duration: float = 0.25) -> None:
        if self._dev is None:
            msg = "Port is not open"
            raise UsbPortError(msg)
        self._dev.ctrl_transfer(
            CTRL_TYPE_CLASS | CTRL_RECIPIENT_INTERFACE,
            self.USB_CDC_CMD_SEND_BREAK,
            wValue=0xFFFF,
            wIndex=0,
        )

    def disconnect(self) -> None:
        if self._iface is not None:
            try:
                usb.util.release_interface(self._dev, self._iface.bInterfaceNumber)
                self._iface = None
            except (usb.core.USBError, AttributeError):
                pass
        if self._dev is not None:
            try:
                usb.util.dispose_resources(self._dev)
            except usb.core.USBError:
                pass

            self._dev = None

    def __del__(self) -> None:
        self.disconnect()

    def close(self) -> None:
        self.disconnect()


class USBLink(BufferedLink):
    def __init__(
        self, vid: Optional[int] = None, pid: Optional[int] = None, serial: Optional[str] = None
    ) -> None:
        super().__init__()
        self._vid = vid
        self._pid = pid
        self._serial = serial
        self._port: Optional[PyUsbCdc] = None

    def _update_timeout(self) -> None:
        # timeout is manually handled in recv/recv_until
        pass

    def connect(self) -> None:
        if self._vid and self._pid:
            self._port = PyUsbCdc(vid=self._vid, pid=self._pid, serial=self._serial, start=False)
        else:
            msg = "Must have vid and pid for usb device"
            raise LinkError(msg)

        if not self._port.open():
            msg = f"Unable to connect to port (vid={self._vid}, pid={self._pid}"
            raise LinkError(msg)

        self._buf = bytearray()
        self.send_break()

    def send_break(self) -> None:
        if self._port is None:
            msg = "Port is not connected"
            raise LinkError(msg)

        self._port.send_break()
        time.sleep(1.0)
        self._port.reset_input_buffer()

    def recv(self, num_bytes: int) -> bytes:
        if self._port is None:
            msg = "Port is not connected"
            raise LinkError(msg)

        t0 = time.monotonic()
        while len(self._buf) < num_bytes:
            if time.monotonic() - t0 > self._timeout:
                msg = "recv timeout"
                raise LinkError(msg)

            try:
                r = bytearray(self._port.read())
            except OSError as e:
                raise LinkError from e
            self._buf.extend(r)

        data = self._buf[:num_bytes]
        self._buf = self._buf[num_bytes:]
        return data

    def recv_until(self, bs: bytes) -> bytes:
        if self._port is None:
            msg = "Port is not connected"
            raise LinkError(msg)

        t0 = time.monotonic()
        while True:
            try:
                i = self._buf.index(bs)
            except ValueError:
                pass
            else:
                break

            if time.monotonic() - t0 > self._timeout:
                msg = "recv timeout"
                raise LinkError(msg)

            try:
                r = bytearray(self._port.read())
            except OSError as e:
                raise LinkError from e
            self._buf.extend(r)

        i += 1
        data = self._buf[:i]
        self._buf = self._buf[i:]

        return data

    def send(self, data: bytes) -> None:
        if self._port is None:
            msg = "Port is not connected"
            raise LinkError(msg)
        self._port.write(data)

    def disconnect(self) -> None:
        if self._port is not None:
            self._port.close()
            self._port = None



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

from enum import Enum
from typing import Optional

import attrs

from acconeer.exptool.app.new.app_model import PluginPresetSpec


@attrs.frozen(kw_only=True)
class PluginPresetBase(PluginPresetSpec):
    name: str = attrs.field()
    description: Optional[str] = attrs.field(default=None)
    preset_id: Enum = attrs.field(default=None)



# Copyright (c) Acconeer AB, 2023-2024
# All rights reserved
from __future__ import annotations

import enum
from typing import List, Optional

import attrs
import numpy as np
import numpy.typing as npt
from attributes_doc import attributes_doc
from scipy.signal import butter

from acconeer.exptool import a121
from acconeer.exptool._core.class_creation.attrs import attrs_ndarray_isclose
from acconeer.exptool.a121.algo import (
    ENVELOPE_FWHM_M,
    AlgoParamEnum,
    AlgoProcessorConfigBase,
    ProcessorBase,
    exponential_smoothing_coefficient,
)
from acconeer.exptool.a121.algo.presence import Processor as PresenceProcessor
from acconeer.exptool.a121.algo.presence import ProcessorConfig as PresenceProcessorConfig
from acconeer.exptool.a121.algo.presence import ProcessorResult as PresenceProcessorResult


class AppState(AlgoParamEnum):
    """Breathing app state."""

    INIT_STATE = enum.auto()

    NO_PRESENCE_DETECTED = enum.auto()
    """No presence detected."""

    INTRA_PRESENCE_DETECTED = enum.auto()
    """Intra presence detected."""

    DETERMINE_DISTANCE_ESTIMATE = enum.auto()
    """Determining distance using presence."""

    ESTIMATE_BREATHING_RATE = enum.auto()
    """Estimating breathing rate."""


@attributes_doc
@attrs.mutable(kw_only=True)
class BreathingProcessorConfig(AlgoProcessorConfigBase):
    lowest_breathing_rate: float = attrs.field(default=6.0)
    """Lowest anticipated breathing rate (breaths per minute)."""

    highest_breathing_rate: float = attrs.field(default=60.0)
    """Highest anticipated breathing rate (breaths per minute)."""

    time_series_length_s: float = attrs.field(default=20.0)
    """Time series length (s)."""

    def _collect_validation_results(
        self, config: a121.SessionConfig
    ) -> List[a121.ValidationResult]:
        validation_results: list[a121.ValidationResult] = []

        if config.sensor_config.frame_rate is None:
            validation_results.append(
                a121.ValidationError(
                    config.sensor_config,
                    "frame_rate",
                    "Must be set",
                )
            )

        return validation_results


@attrs.frozen(kw_only=True)
class BreathingProcessorExtraResult:
    psd: npt.NDArray[np.float64] = attrs.field(eq=attrs_ndarray_isclose)
    frequencies: npt.NDArray[np.float64] = attrs.field(eq=attrs_ndarray_isclose)
    breathing_motion: npt.NDArray[np.float64] = attrs.field(eq=attrs_ndarray_isclose)
    time_vector: npt.NDArray[np.float64] = attrs.field(eq=attrs_ndarray_isclose)
    breathing_rate_history: npt.NDArray[np.float64] = attrs.field(eq=attrs_ndarray_isclose)
    all_breathing_rate_history: npt.NDArray[np.float64] = attrs.field(eq=attrs_ndarray_isclose)


@attrs.mutable(kw_only=True)
class BreathingProcessorResult:
    breathing_rate: Optional[float] = attrs.field(default=None)
    """Estimated breathing rate. Breaths per minute."""

    extra_result: BreathingProcessorExtraResult
    """Extra result, only used for visualization."""


class BreathingProcessor(ProcessorBase[BreathingProcessorResult]):
    """Breathing rate processor."""

    SECONDS_IN_MINUTE: float = 60.0
    HISTORY_S = SECONDS_IN_MINUTE * 2.0

    # Type declarations
    start_point: int
    end_point: int
    sparse_iq_buffer: npt.NDArray[np.complex128]
    filt_sparse_iq_buffer: npt.NDArray[np.complex128]
    angle_buffer: npt.NDArray[np.float64]
    filt_angle_buffer: npt.NDArray[np.float64]
    breathing_motion_buffer: npt.NDArray[np.float64]
    breathing_rate_history: npt.NDArray[np.float64]
    all_breathing_rate_history: npt.NDArray[np.float64]

    start_time: float
    init_counter: int
    point_counter: int
    prev_angle: Optional[float]
    lp_filt_ampl: Optional[float]
    angle_unwrapped: npt.NDArray[np.float64]

    def __init__(
        self,
        *,
        sensor_config: a121.SensorConfig,
        processor_config: BreathingProcessorConfig,
    ):
        assert sensor_config.frame_rate is not None

        lowest_breathing_rate_hz = processor_config.lowest_breathing_rate / self.SECONDS_IN_MINUTE
        highest_breathing_rate_hz = (
            processor_config.highest_breathing_rate / self.SECONDS_IN_MINUTE
        )
        self.frame_rate = sensor_config.frame_rate
        self.time_series_length_s = processor_config.time_series_length_s
        self.time_series_length = int(self.time_series_length_s * self.frame_rate)
        self.padded_time_series_length = 2 ** (int(np.log2(self.time_series_length)) + 1)
        self.analysis_overlap = int(self.time_series_length / 2)
        self.num_points = sensor_config.num_points

        # Filter coefficients.
        self.b_static, self.a_static = butter(
            N=2, Wn=lowest_breathing_rate_hz, btype="lowpass", fs=self.frame_rate
        )
        self.b_angle, self.a_angle = butter(
            N=2,
            Wn=[lowest_breathing_rate_hz, highest_breathing_rate_hz],
            btype="bandpass",
            fs=self.frame_rate,
        )
        self.sf = exponential_smoothing_coefficient(self.frame_rate, self.time_series_length_s)

        # PSD frequency vector.
        self.frequencies = np.fft.rfftfreq(self.padded_time_series_length, 1 / self.frame_rate)
        self.time_vector = np.linspace(-self.HISTORY_S, 0, int(self.frame_rate * self.HISTORY_S))

        self.reinitialize_processor(0, self.num_points)

    def process(self, result: a121.Result) -> BreathingProcessorResult:
        frame = result.frame[:, self.start_point : self.end_point]
        mean_sweep = frame.mean(axis=0)

        # Estimate static component
        self.sparse_iq_buffer = np.roll(self.sparse_iq_buffer, shift=1, axis=0)
        self.sparse_iq_buffer[0] = mean_sweep

        filt_sparse_iq = -np.sum(
            self.a_static[1:][:, np.newaxis] * self.filt_sparse_iq_buffer, axis=0
        ) + np.sum(self.b_static[:, np.newaxis] * self.sparse_iq_buffer, axis=0)

        self.filt_sparse_iq_buffer = np.roll(self.filt_sparse_iq_buffer, shift=1, axis=0)
        self.filt_sparse_iq_buffer[0] = filt_sparse_iq

        # Remove static components by subtracting the estimated mean.
        zm_sweep = mean_sweep - filt_sparse_iq
        angle = np.angle(zm_sweep)

        if self.prev_angle is None or self.lp_filt_ampl is None:
            self.prev_angle = angle
            self.lp_filt_ampl = np.abs(zm_sweep)

        # Filter the amplitude(used when weighting psds from the measured distances).
        assert self.lp_filt_ampl is not None
        self.lp_filt_ampl = self.sf * self.lp_filt_ampl + (1 - self.sf) * np.abs(zm_sweep)

        # Unwrap the angles.
        angle_diff = angle - self.prev_angle
        angle_diff[np.pi < angle_diff] -= 2 * np.pi
        angle_diff[angle_diff < -np.pi] += 2 * np.pi
        self.angle_unwrapped = self.angle_unwrapped + angle_diff
        self.angle_buffer = np.roll(self.angle_buffer, shift=1, axis=0)
        self.angle_buffer[0] = self.angle_unwrapped
        self.prev_angle = angle

        # Bandpass filter angles.
        filt_angle = -np.sum(
            self.a_angle[1:][:, np.newaxis] * self.filt_angle_buffer, axis=0
        ) + np.sum(self.b_angle[:, np.newaxis] * self.angle_buffer, axis=0)
        self.filt_angle_buffer = np.roll(self.filt_angle_buffer, shift=1, axis=0)
        self.filt_angle_buffer[0] = filt_angle

        # Add filtered angle to breathing motion fifo buffer.
        self.breathing_motion_buffer = np.roll(self.breathing_motion_buffer, shift=-1, axis=0)
        self.breathing_motion_buffer[-1] = filt_angle

        # Calculate psd of signal.
        windowed_breathing_motion_buffer = (
            self.breathing_motion_buffer * np.hamming(self.time_series_length)[:, np.newaxis]
        )
        psd = np.fft.rfft(
            windowed_breathing_motion_buffer, axis=0, n=self.padded_time_series_length
        )
        # Omit **2 to reduce processing as it does not alter the result.
        psd = np.abs(psd)
        assert self.lp_filt_ampl is not None
        psd_weighted = np.sum(psd * self.lp_filt_ampl, axis=1) / np.sum(self.lp_filt_ampl)

        # Interpolate around peak to gain better resolution.
        # Wait until data of a full time series is available.
        peak_loc = np.argmax(psd_weighted)
        if peak_loc != 0 and self.time_series_length < self.init_counter:
            estimated_frequency = self._peak_interpolation(
                psd_weighted[peak_loc - 1 : peak_loc + 2],
                self.frequencies[peak_loc - 1 : peak_loc + 2],
            )
            estimated_breathing_rate = estimated_frequency * self.SECONDS_IN_MINUTE
        else:
            self.init_counter += 1
            estimated_breathing_rate = None

        # Shift breathing rate history and add latest estimate.
        self.all_breathing_rate_history = np.roll(self.all_breathing_rate_history, shift=-1)
        self.all_breathing_rate_history[-1] = estimated_breathing_rate
        self.breathing_rate_history = np.roll(self.breathing_rate_history, shift=-1)

        # Report breathing rate if enough time has elapsed since last estimate.
        if self.time_series_length - self.analysis_overlap <= self.point_counter:
            self.breathing_rate_history[-1] = estimated_breathing_rate
            self.point_counter = 0
        else:
            self.breathing_rate_history[-1] = np.nan
            self.point_counter += 1

        # Prepare extra result, used for plotting.
        extra_result = BreathingProcessorExtraResult(
            psd=psd_weighted,
            frequencies=self.frequencies,
            breathing_motion=self.breathing_motion_buffer[:, self.center_distance_idx],
            time_vector=self.time_vector,
            all_breathing_rate_history=self.all_breathing_rate_history,
            breathing_rate_history=self.breathing_rate_history,
        )

        return BreathingProcessorResult(
            breathing_rate=estimated_breathing_rate, extra_result=extra_result
        )

    def reinitialize_processor(self, start_point: int, end_point: int) -> None:
        self.start_point = start_point
        self.end_point = end_point

        num_points_to_analyze = self.end_point - self.start_point
        self.center_distance_idx = int(num_points_to_analyze / 2)

        # Memory of IIR filters.
        self.sparse_iq_buffer = np.zeros(
            shape=(self.b_static.size, num_points_to_analyze), dtype="complex128"
        )
        self.filt_sparse_iq_buffer = np.zeros(
            shape=(self.a_static.size - 1, num_points_to_analyze), dtype="complex128"
        )

        self.angle_buffer = np.zeros(shape=(self.b_angle.size, num_points_to_analyze))
        self.filt_angle_buffer = np.zeros(shape=(self.a_angle.size - 1, num_points_to_analyze))

        # Memory for breathing motion time series.
        self.breathing_motion_buffer = np.zeros(
            shape=(self.time_series_length, num_points_to_analyze)
        )

        # State variables.
        self.init_counter = 0
        self.prev_angle = None
        self.lp_filt_ampl = None
        self.point_counter = 0
        self.angle_unwrapped = np.zeros(shape=num_points_to_analyze)

        # Memory for breathing rate history.
        self.breathing_rate_history = np.full(
            shape=int(self.frame_rate * self.HISTORY_S), fill_value=np.nan
        )
        self.all_breathing_rate_history = np.full(
            shape=int(self.frame_rate * self.HISTORY_S), fill_value=np.nan
        )

    @staticmethod
    def _peak_interpolation(y: npt.NDArray[np.float64], x: npt.NDArray[np.float64]) -> float:
        """Quadratic interpolation of three points.

        Derivation:
        https://math.stackexchange.com/questions/680646/get-polynomial-function-from-3-points
        """
        a = (x[0] * (y[2] - y[1]) + x[1] * (y[0] - y[2]) + x[2] * (y[1] - y[0])) / (
            (x[0] - x[1]) * (x[0] - x[2]) * (x[1] - x[2])
        )
        b = (y[1] - y[0]) / (x[1] - x[0]) - a * (x[0] + x[1])
        peak_loc = -b / (2 * a)
        return float(peak_loc)


def get_presence_config() -> PresenceProcessorConfig:
    presence_config = PresenceProcessorConfig()
    presence_config.intra_detection_threshold = 4.0
    return presence_config


class ProcessorConfig(AlgoProcessorConfigBase):
    num_distances_to_analyze: int = attrs.field(default=3)
    """Indicates the number of distance to analyzed, centered around the distance where presence
    is detected."""

    distance_determination_duration: float = attrs.field(default=5.0)
    """Time for the presence processor to determine distance to presence."""

    use_presence_processor: bool = attrs.field(default=True)
    """If True, use the presence processor to determine distance to subject."""

    breathing_config: BreathingProcessorConfig = attrs.field(factory=BreathingProcessorConfig)

    presence_config: PresenceProcessorConfig = attrs.field(factory=get_presence_config)

    def _collect_validation_results(
        self, config: a121.SessionConfig
    ) -> List[a121.ValidationResult]:
        validation_results: list[a121.ValidationResult] = []

        if config.sensor_config.frame_rate is None:
            validation_results.append(
                a121.ValidationError(
                    config.sensor_config,
                    "frame_rate",
                    "Must be set",
                )
            )

        return validation_results


@attrs.mutable(kw_only=True)
class ProcessorResult:
    app_state: AppState
    """Application state."""

    distances_being_analyzed: Optional[tuple[int, int]] = None
    """Range where breathing is being analyzed."""

    presence_result: PresenceProcessorResult
    """Presence processor result."""

    breathing_result: Optional[BreathingProcessorResult] = attrs.field(default=None)
    """Breathing processor result."""


class Processor(ProcessorBase[ProcessorResult]):
    """Breathing rate super-processor.

    Handles execution of the presence processor and the breathing processor.
    """

    def __init__(
        self,
        *,
        sensor_config: a121.SensorConfig,
        processor_config: ProcessorConfig,
        metadata: a121.Metadata,
    ):
        assert sensor_config.frame_rate is not None

        self.sensor_config = sensor_config
        self.metadata = metadata

        self.presence_processor = PresenceProcessor(
            sensor_config=sensor_config,
            metadata=self.metadata,
            processor_config=processor_config.presence_config,
        )

        self.breathing_processor = BreathingProcessor(
            sensor_config=sensor_config,
            processor_config=processor_config.breathing_config,
        )

        self.breathing_processor.reinitialize_processor(0, sensor_config.num_points)

        # Parameters
        self.intra_detection_threshold = processor_config.presence_config.intra_detection_threshold
        self.use_presence_processor = processor_config.use_presence_processor
        self.presence_threshold = ENVELOPE_FWHM_M[sensor_config.profile] * 2.0
        self.num_distances_to_analyze_half_width = int(
            processor_config.num_distances_to_analyze / 2.0
        )
        self.distance_determination_duration = (
            processor_config.distance_determination_duration * sensor_config.frame_rate
        )
        self.sf = exponential_smoothing_coefficient(
            sensor_config.frame_rate, processor_config.distance_determination_duration / 4.0
        )
        self.num_points = sensor_config.num_points

        # State variables
        self.presence_distance_has_init = False
        self.distance_determination_counter = 0
        self.app_state: AppState = AppState.INIT_STATE
        self.prev_app_state = AppState.INIT_STATE
        self.base_presence_distance: Optional[float] = None

    def process(self, result: a121.Result) -> ProcessorResult:
        presence_result = self.presence_processor.process(result=result)

        self._determine_app_state(presence_result)

        self._update_presence_distance(presence_result.presence_distance)

        processor_result = ProcessorResult(
            app_state=self.app_state, presence_result=presence_result
        )

        self._perform_action_based_on_app_state(result, processor_result)

        self.prev_app_state = self.app_state

        return processor_result

    def _determine_app_state(self, presence_result: PresenceProcessorResult) -> None:
        if not presence_result.presence_detected:
            self.app_state = AppState.NO_PRESENCE_DETECTED

        elif self.intra_detection_threshold < presence_result.intra_presence_score:
            self.app_state = AppState.INTRA_PRESENCE_DETECTED

        elif self.base_presence_distance is None and self.use_presence_processor:
            self.app_state = AppState.DETERMINE_DISTANCE_ESTIMATE

        elif (
            not self.use_presence_processor
            or self.distance_determination_duration <= self.distance_determination_counter
        ):
            self.app_state = AppState.ESTIMATE_BREATHING_RATE

        else:
            # Do not change app state
            pass

    def _update_presence_distance(self, presence_distance: float) -> None:
        # Calculate and filter presence distance.
        if not self.presence_distance_has_init:
            self.presence_distance = presence_distance
            self.presence_distance_has_init = True

        assert self.presence_distance is not None
        self.presence_distance = self.presence_distance * self.sf + presence_distance * (
            1 - self.sf
        )

        # Determine if presence location has changed.
        if self.base_presence_distance is not None and self.presence_threshold < np.abs(
            self.base_presence_distance - self.presence_distance
        ):
            self.base_presence_distance = None

    def _perform_action_based_on_app_state(
        self, result: a121.Result, processor_result: ProcessorResult
    ) -> None:
        # Perform action based on app state
        if (
            self.app_state == AppState.INTRA_PRESENCE_DETECTED
            or self.app_state == AppState.NO_PRESENCE_DETECTED
        ):
            self.base_presence_distance = None

        elif self.app_state == AppState.DETERMINE_DISTANCE_ESTIMATE:
            if self.app_state != self.prev_app_state:
                self.distance_determination_counter = 0
            else:
                self.distance_determination_counter += 1
                self.base_presence_distance = self.presence_distance

        elif self.app_state == AppState.ESTIMATE_BREATHING_RATE:
            if self.app_state != self.prev_app_state:
                if self.use_presence_processor:
                    center_idx = self._base_presence_distance_to_point()
                    start_point = max(0, center_idx - self.num_distances_to_analyze_half_width)
                    end_point = min(
                        center_idx + self.num_distances_to_analyze_half_width + 1,
                        self.num_points,
                    )
                else:
                    start_point = 0
                    end_point = self.num_points

                self.breathing_processor.reinitialize_processor(
                    start_point=start_point,
                    end_point=end_point,
                )

            processor_result.distances_being_analyzed = (
                self.breathing_processor.start_point,
                self.breathing_processor.end_point,
            )
            processor_result.breathing_result = self.breathing_processor.process(result)

        else:
            msg = "Invalid app state"
            raise NotImplementedError(msg)

    def _base_presence_distance_to_point(self) -> int:
        """Calculates the closest point of a distance in meters"""
        assert self.base_presence_distance is not None
        assert not isinstance(self.metadata, list)

        measured_points = (
            self.sensor_config.start_point
            + np.arange(self.sensor_config.num_points) * self.sensor_config.step_length
        )
        center_point = int(self.base_presence_distance / self.metadata.base_step_length_m)
        return int(np.argmin(np.abs(measured_points - center_point)))



# Copyright (c) Acconeer AB, 2023-2024
# All rights reserved

from __future__ import annotations

import types
import typing as t

import typing_extensions as te
from typing_extensions import Concatenate as C


#              name kwargs
Task = t.Tuple[str, t.Dict[str, t.Any]]

_TaskSender = t.Callable[[Task], t.Any]
_R = t.TypeVar("_R")
_T = t.TypeVar("_T")
_P = te.ParamSpec("_P")


class _TaskDescriptor(t.Generic[_P, _R]):
    def __init__(self, method: t.Callable[C[t.Any, _P], _R]) -> None:
        self._method = method
        self._task_name: t.Optional[str] = None
        setattr(self._method, "is_task", True)

    def __set_name__(self, owner: t.Type[t.Any], name: str) -> None:
        self._task_name = name

    @t.overload
    def __get__(self, obj: None, objtype: t.Any = ...) -> te.Self: ...

    @t.overload
    def __get__(self, obj: t.Any, objtype: t.Any = ...) -> t.Callable[_P, _R]: ...

    def __get__(
        self, obj: t.Optional[t.Any], objtype: t.Any = None
    ) -> t.Union[te.Self, t.Callable[_P, _R]]:
        """If called with an instance, bind and return the wrapped function"""
        if obj is None:
            return self
        else:
            bound_method = types.MethodType(self._method, obj)
            return bound_method

    def rpc(self, task_sender: _TaskSender, *args: _P.args, **kwargs: _P.kwargs) -> None:
        if self._task_name is None:
            raise RuntimeError

        if args != ():
            msg = "An RPC must be called with kwargs only."
            raise ValueError(msg)

        task_sender((self._task_name, kwargs))


def is_task(m: t.Callable[C[t.Any, _P], _R]) -> _TaskDescriptor[_P, _R]:
    """
    Task method decorator

    The decorated method gets wrapped in a descriptor and works normally on the surface:

        class Foo:
            @is_task
            def my_task(self, *, n: int) -> None:
                ...

        Foo().my_task(n=5)

    The decorator also provides a rpc member "of the method":

        class FooView:
            def on_some_event(self):
                Foo.my_task.rpc(<task_sender>, n=5)

    Which will construct and send a Task with <task_sender>.

    <task_sender> should be AppModel.put_task in most cases.
    """
    return _TaskDescriptor(m)


def _has_task_marker(obj: object) -> bool:
    return getattr(obj, "is_task", False)


def get_task(obj: t.Any, name: str) -> t.Optional[t.Callable[..., t.Any]]:
    """
    Gets the task method "<name>" of the obj.
    If obj does not have a task named "<name>", None is returned.
    """
    task_method = getattr(obj, name, None)
    if _has_task_marker(task_method):
        return task_method
    else:
        return None


def get_task_names(obj: object) -> list[str]:
    """
    Gets the names of all tasks defined on 'obj'
    """
    return [
        name
        for name in dir(obj)
        if _has_task_marker(getattr(obj, name, None)) and not name.startswith("_task_method_")
    ]



# Copyright (c) Acconeer AB, 2023-2025
# All rights reserved

from __future__ import annotations

import copy
import enum
import warnings
from typing import Any, Dict, List, Optional, Tuple

import attrs
import h5py
import numpy as np
import numpy.typing as npt
from attributes_doc import attributes_doc

from acconeer.exptool import a121, opser
from acconeer.exptool._core.class_creation.attrs import attrs_optional_ndarray_isclose
from acconeer.exptool.a121._h5_utils import _create_h5_string_dataset
from acconeer.exptool.a121.algo import (
    APPROX_BASE_STEP_LENGTH_M,
    ENVELOPE_FWHM_M,
    AlgoBase,
    AlgoConfigBase,
    Controller,
)
from acconeer.exptool.a121.algo._utils import get_distances_m
from acconeer.exptool.a121.algo.parking._processors import (
    ObstructionProcessor,
    ObstructionProcessorConfig,
    Processor,
    ProcessorConfig,
)


class DetailedStatus(enum.Enum):
    OK = enum.auto()
    CALIBRATION_MISSING = enum.auto()
    CONFIG_MISMATCH = enum.auto()
    OBSTRUCTION_CALIBRATION_MISSING = enum.auto()


@attrs.frozen(kw_only=True)
class RefAppStatus:
    ref_app_state: DetailedStatus
    ready_to_start: bool


@attrs.mutable(kw_only=True)
class CalibrationData(AlgoBase):
    noise_frames: Optional[npt.NDArray[np.complex128]] = attrs.field(
        default=None, eq=attrs_optional_ndarray_isclose
    )
    obs_tx_off_frames: Optional[npt.NDArray[np.complex128]] = attrs.field(
        default=None, eq=attrs_optional_ndarray_isclose
    )
    obs_frames: Optional[npt.NDArray[np.complex128]] = attrs.field(
        default=None, eq=attrs_optional_ndarray_isclose
    )
    temperatures: Optional[List[int]] = attrs.field(default=None)


@attrs.mutable(kw_only=True)
class RefAppContext(AlgoBase):
    noise_level: float = attrs.field(default=1.0)
    obstruction_center: npt.NDArray[np.float64] = attrs.field(default=np.array([0.0, 0.0]))
    calibration_temperature: float = attrs.field(default=0.0)
    calibration_done: bool = attrs.field(default=False)
    calibration_sensor_config: a121.SensorConfig = attrs.field(factory=a121.SensorConfig)
    obstruction_calibration_done: bool = attrs.field(default=False)
    obstruction_noise_level: float = attrs.field(default=0.0)
    obstruction_sensor_config: a121.SensorConfig = attrs.field(factory=a121.SensorConfig)
    calibration_data: CalibrationData = attrs.field(factory=CalibrationData)


@attributes_doc
@attrs.mutable(kw_only=True)
class RefAppConfig(AlgoConfigBase):
    range_start_m: float = attrs.field(default=0.3)
    """Start of measurement range (m)."""

    range_end_m: float = attrs.field(default=1.5)
    """End of measurement range (m)."""

    hwaas: int = attrs.field(default=32)
    """HWAAS. This parameter has an impact on the overall signal to noise ratio (SNR), low values
    will yield a noisy signal and poorer performance, since the signature will vary more due to
    noise variations."""

    profile: a121.Profile = attrs.field(default=a121.Profile.PROFILE_2, converter=a121.Profile)
    """Profile. This parameter determines the length of the pulse. If the range is within the so
    called "direct leakage", the performance of the algorithm will deteriorate.
    For close ranges (i.e. ground mounted applications), it is recommended to use profile 1 or 2.
    For longer ranges, higher profiles should be used. The limitations that the range puts on
    profile choices is implemented in the Exploration tool GUI and can be experimented
    with there."""

    update_rate: float = attrs.field(default=10.0)
    """Update rate."""

    queue_length_n: int = attrs.field(default=5)
    """Number of consecutive samples stored."""

    amplitude_threshold: float = attrs.field(default=15.0)
    """Threshold level in times the noise level to trigger."""

    weighted_distance_threshold_m: float = attrs.field(default=0.1)
    """Max distance between two signatures to consider samples similar."""

    signature_similarity_threshold: Optional[float] = attrs.field(default=60.0)
    """How large fraction of signatures have to be within distance threshold to detect car."""

    obstruction_detection: bool = attrs.field(default=False)
    """If obstruction detection should be active. This costs some extra battery power."""

    obstruction_distance_threshold: float = attrs.field(default=0.06)
    """Determines how large fraction of each dimension the obstruction measurement can vary without
    considering the sensor obstructed. E.g. if the threshold is set to 0.06 and the range is
    between 0.02 m and 0.05 m. The weighted distance part of the signature
    can vary (0.05 m - 0.02 m) * 0.06 = 0.0018 m.
    """

    obstruction_start_m: float = attrs.field(default=0.03)
    """Start of obstruction detection range. The detection range should not intersect with the detection range for the main functionality."""

    obstruction_end_m: float = attrs.field(default=0.05)
    """End of obstruction detection range. The detection range should not intersect with the detection range for the main functionality."""

    @classmethod
    def from_dict(cls, d: dict[str, Any]) -> RefAppConfig:
        return cls(**d)

    def _collect_validation_results(self) -> list[a121.ValidationResult]:
        validation_results: list[a121.ValidationResult] = []
        fwhm = ENVELOPE_FWHM_M[self.profile]

        if self.range_start_m < 2 * fwhm:
            validation_results.append(
                a121.ValidationError(
                    self,
                    "range_start_m",
                    "Start point inside direct leakage",
                )
            )

        if self.range_end_m < self.range_start_m:
            validation_results.append(
                a121.ValidationError(
                    self,
                    "range_end_m",
                    "End point cannot be behind start point",
                )
            )

        if self.range_end_m > 5.0:
            validation_results.append(
                a121.ValidationError(
                    self,
                    "range_end_m",
                    "End point must be within 5m, see documentation for details.",
                )
            )

        if self.obstruction_detection and (self.obstruction_end_m < self.obstruction_start_m):
            validation_results.append(
                a121.ValidationError(
                    self,
                    "obstruction_end_m",
                    "Obstruction end point cannot be behind start point",
                )
            )

        return validation_results


@attrs.frozen(kw_only=True)
class RefAppExtraResult:
    signature_history: npt.NDArray[np.float64] = attrs.field()
    """Array containing queue_length_n last signatures."""

    parking_data: npt.NDArray[np.float64] = attrs.field()
    """The scaled amplitude array used to calculate the last signature."""

    closest_object_dist: float = attrs.field(default=0.0)
    """Location of closest signature in cluster. Only availible if car is detected."""

    obstruction_data: Optional[npt.NDArray[np.float64]] = attrs.field(default=None)
    """Array with obstruction amplitudes (usually direct leakage)."""

    obstruction_signature: npt.NDArray[np.float64] = attrs.field(default=(0.0, 0.0))
    """Signature of amplitudes used for the obstruction detection."""

    obstruction_center: npt.NDArray[np.float64] = attrs.field(default=(0.0, 0.0))
    """Signature of amplitudes used in calibration."""

    obstruction_distance: float = attrs.field(default=0.0)
    """Obstruction distance threshold, same as in configuration."""


@attrs.frozen(kw_only=True)
class RefAppResult:
    car_detected: bool = attrs.field(default=False)
    """Boolean indicating if a car is present."""

    obstruction_detected: bool = attrs.field(default=False)
    """Boolean indicating whether something is obstructing the sensor."""

    extra_result: RefAppExtraResult = attrs.field()
    """Extra information for plotting."""


class RefApp(Controller[RefAppConfig, RefAppResult]):
    def __init__(
        self,
        *,
        client: a121.Client,
        sensor_id: int,
        ref_app_config: RefAppConfig,
        context: RefAppContext = RefAppContext(),
    ) -> None:
        super().__init__(client=client, config=ref_app_config)

        self.all_sensor_configs = get_sensor_configs(ref_app_config)
        self.sensor_config = self.all_sensor_configs["full_config"]
        self.sensor_id = sensor_id
        self.n_samples_to_calibrate = 3  # number of samples used for calibration.

        if ref_app_config.obstruction_detection:
            self.obstruction_config = self.all_sensor_configs["obstruction_config"]

        self.session_config = a121.SessionConfig(
            {self.sensor_id: self.sensor_config},
            update_rate=ref_app_config.update_rate,
        )

        if ref_app_config.signature_similarity_threshold is not None:
            # input is a percentage, divide to get useful value
            signature_similarity_threshold = ref_app_config.signature_similarity_threshold / 100.0
        else:
            signature_similarity_threshold = 0.6

        self.processor_config = ProcessorConfig(
            queue_length=ref_app_config.queue_length_n,
            amplitude_threshold=ref_app_config.amplitude_threshold,
            weighted_distance_threshold_m=ref_app_config.weighted_distance_threshold_m,
            signature_similarity_threshold=signature_similarity_threshold,
        )
        self.ref_app_config = ref_app_config
        self.sensor_id = sensor_id

        self.context = context

        self.noise_level = self.context.noise_level  # noise level estimated at calibration
        self.calibration_temperature = self.context.calibration_temperature

        if ref_app_config.obstruction_detection:
            self.obstruction_center = self.context.obstruction_center
            self.obstruction_noise_level = self.context.obstruction_noise_level

        self.calibration_done = False
        self.started = False

    def start(
        self,
        recorder: Optional[a121.Recorder] = None,
        algo_group: Optional[h5py.Group] = None,
    ) -> None:
        if self.started:
            msg = "Already started"
            raise RuntimeError(msg)

        self.metadata = self.client.setup_session(self.session_config)
        assert isinstance(self.metadata, a121.Metadata)

        if recorder is not None:
            if isinstance(recorder, a121.H5Recorder):
                if algo_group is None:
                    algo_group = recorder.require_algo_group("parking")
                _record_algo_data(
                    algo_group,
                    self.sensor_id,
                    self.config,
                    self.context,
                )
            else:
                # Should never happen as we currently only have the H5Recorder
                warnings.warn("Will not save algo data")

            self.client.attach_recorder(recorder)

        base_sensor_config = self.all_sensor_configs["full_config"]
        base_sensor_config = a121.SensorConfig(subsweeps=[base_sensor_config.subsweeps[0]])

        self.base_processor = Processor(
            sensor_config=base_sensor_config,
            processor_config=self.processor_config,
            metadata=self.metadata,
            noise_estimate=self.noise_level,
            noise_estimate_temperature=self.calibration_temperature,
        )

        if self.ref_app_config.obstruction_detection:
            obstruction_sensor_config = self.all_sensor_configs["obstruction_config"]
            obstruction_distance_threshold = self.ref_app_config.obstruction_distance_threshold
            obstruction_processor_config = ObstructionProcessorConfig(
                distance_threshold=obstruction_distance_threshold
            )
            self.obs_distances = get_distances_m(obstruction_sensor_config, self.metadata)

            self.obstruction_processor = ObstructionProcessor(
                sensor_config=obstruction_sensor_config,
                processor_config=obstruction_processor_config,
                metadata=self.metadata,
                update_rate=self.ref_app_config.update_rate,
                calibration_center=self.obstruction_center,
                calibration_noise_mean=self.obstruction_noise_level,
                calibration_temperature=self.calibration_temperature,
            )

        self.client.start_session()

        self.started = True

    def calibrate_ref_app(self) -> None:
        obstruction_detection = self.ref_app_config.obstruction_detection

        sensor_configs = get_sensor_configs(self.ref_app_config)
        sensor_config = sensor_configs["full_config"]

        # Turn off tx in all
        for subsweep in sensor_config.subsweeps:
            subsweep.enable_tx = False

        if obstruction_detection:
            # We need two calibration values for obstruction, one to determine the signature and one to determine the noise level.

            obstruction_config = copy.deepcopy(sensor_config)
            obstruction_config.subsweeps[1].enable_tx = True

            groups = [
                {self.sensor_id: sensor_config},  # This is for parking noise
                {self.sensor_id: sensor_config},  # This is for obstruction noise
                {self.sensor_id: obstruction_config},  # This is for obstruction center
            ]
            session_config = a121.SessionConfig(
                groups,
                update_rate=self.n_samples_to_calibrate,
            )
        else:
            session_config = a121.SessionConfig(
                sensor_config,
                update_rate=self.n_samples_to_calibrate,
            )

        metadata = self.client.setup_session(session_config)

        if obstruction_detection:
            assert isinstance(metadata, list)
            obstruction_config = sensor_configs["obstruction_config"]
            obstruction_metadata = metadata[2][self.sensor_id]
            obs_distances = get_distances_m(obstruction_config, obstruction_metadata)

        noise_levels = []
        obs_noise_levels = []
        signatures = []
        temperatures = []

        noise_frames = []
        obs_tx_off_frames = []
        obs_frames = []

        self.client.start_session()
        for i in range(self.n_samples_to_calibrate):
            base_result = self.client.get_next()
            if obstruction_detection:
                assert isinstance(base_result, list)
                main_result = base_result[0][self.sensor_id]
                main_frame = main_result.subframes[0]
                obs_tx_off_frame = base_result[1][self.sensor_id].subframes[1]
                obs_frame = base_result[2][self.sensor_id].subframes[1]
                obs_noise_level = ObstructionProcessor.get_noise_level(obs_tx_off_frame)
                obs_noise_levels.append(obs_noise_level)
                signature, _ = ObstructionProcessor.get_signature(
                    obs_frame, obs_noise_level, obs_distances
                )
                signatures.append(signature)

                obs_tx_off_frames.append(obs_tx_off_frame)
                obs_frames.append(obs_frame)
            else:
                assert isinstance(base_result, a121.Result)
                main_result = base_result
                main_frame = main_result.frame
            temperature = main_result.temperature
            temperatures.append(temperature)

            noise_level = Processor.process_noise_frame(main_frame)
            noise_levels.append(noise_level)
            noise_frames.append(main_frame)

        if obstruction_detection:
            self.obstruction_center = np.mean(signatures, axis=0)
            self.context.obstruction_noise_level = float(np.mean(obs_noise_levels))
            self.context.obstruction_center = self.obstruction_center
            self.context.obstruction_calibration_done = True
            self.context.obstruction_sensor_config = obstruction_config
        else:
            self.context.obstruction_calibration_done = False

        self.client.stop_session()

        self.calibration_temperature = float(np.median(temperatures))
        self.noise_level = float(np.mean(noise_levels))

        # Update the context
        for subsweep in sensor_config.subsweeps:
            subsweep.enable_tx = True
        self.context.noise_level = self.noise_level
        self.context.calibration_temperature = self.calibration_temperature
        self.context.calibration_sensor_config = sensor_config
        self.context.calibration_done = True

        self.context.calibration_data.noise_frames = np.array(noise_frames)
        self.context.calibration_data.obs_tx_off_frames = np.array(obs_tx_off_frames)
        self.context.calibration_data.obs_frames = np.array(obs_frames)
        self.context.calibration_data.temperatures = temperatures

    def get_next(self) -> RefAppResult:
        if not self.started:
            msg = "Not started"
            raise RuntimeError(msg)

        result = self.client.get_next()

        assert isinstance(result, a121.Result)
        if self.ref_app_config.obstruction_detection:
            base_frame = result.subframes[0]
            obs_frame = result.subframes[1]
        else:
            base_frame = result.subframes[0]

        temperature = result.temperature

        processor_result = self.base_processor.process(base_frame, temperature)

        if self.ref_app_config.obstruction_detection:
            obstruction_result = self.obstruction_processor.process(obs_frame, temperature)

            obstruction_found = obstruction_result.obstruction_found

            extra_result = RefAppExtraResult(
                signature_history=processor_result.extra_result.signature_history,
                parking_data=processor_result.extra_result.parking_data,
                closest_object_dist=processor_result.extra_result.closest_observation,
                obstruction_data=obstruction_result.extra_result.obstruction_data,
                obstruction_signature=obstruction_result.extra_result.obstruction_signature,
                obstruction_center=obstruction_result.extra_result.obstruction_center,
                obstruction_distance=obstruction_result.extra_result.obstruction_distance,
            )
        else:
            obstruction_found = False
            extra_result = RefAppExtraResult(
                signature_history=processor_result.extra_result.signature_history,
                parking_data=processor_result.extra_result.parking_data,
                closest_object_dist=processor_result.extra_result.closest_observation,
            )

        ref_app_result = RefAppResult(
            car_detected=processor_result.car_detected,
            obstruction_detected=obstruction_found,
            extra_result=extra_result,
        )

        return ref_app_result

    def stop(self) -> Any:
        if not self.started:
            msg = "Already stopped"
            raise RuntimeError(msg)

        self.client.stop_session()
        recorder = self.client.detach_recorder()
        if recorder is None:
            recorder_result = None
        else:
            recorder_result = recorder.close()

        self.started = False

        return recorder_result

    @classmethod
    def get_ref_app_status(
        cls, ref_app_config: RefAppConfig, context: RefAppContext
    ) -> RefAppStatus:
        calibration_done = context.calibration_done

        sensor_configs = get_sensor_configs(ref_app_config)
        sensor_config = sensor_configs["full_config"]

        config_match = context.calibration_sensor_config == sensor_config

        if ref_app_config.obstruction_detection:
            obstruction_conf = sensor_configs["obstruction_config"]
            obstruction_config_match = obstruction_conf == context.obstruction_sensor_config
            obstruction_ready = context.obstruction_calibration_done and obstruction_config_match
        else:
            obstruction_ready = True

        if not calibration_done:
            ret = RefAppStatus(
                ready_to_start=False,
                ref_app_state=DetailedStatus.CALIBRATION_MISSING,
            )
        elif not config_match:
            ret = RefAppStatus(
                ready_to_start=False,
                ref_app_state=DetailedStatus.CONFIG_MISMATCH,
            )
        elif not obstruction_ready:
            ret = RefAppStatus(
                ready_to_start=False,
                ref_app_state=DetailedStatus.OBSTRUCTION_CALIBRATION_MISSING,
            )
        else:
            ret = RefAppStatus(ready_to_start=True, ref_app_state=DetailedStatus.OK)

        return ret


def _record_algo_data(
    algo_group: h5py.Group,
    sensor_id: int,
    ref_app_config: RefAppConfig,
    context: RefAppContext,
) -> None:
    algo_group.create_dataset(
        "ref_app_sensor_id",
        data=sensor_id,
        track_times=False,
    )
    _create_h5_string_dataset(algo_group, "ref_app_config", ref_app_config.to_json())

    context_group = algo_group.create_group("context")
    opser.serialize(context, context_group)


def _load_algo_data(algo_group: h5py.Group) -> Tuple[int, RefAppConfig, RefAppContext]:
    sensor_id = int(algo_group["ref_app_sensor_id"][()])

    config = RefAppConfig.from_json(algo_group["ref_app_config"][()])

    context_group = algo_group["context"]
    context = opser.deserialize(context_group, RefAppContext)

    return sensor_id, config, context


def get_obstruction_subsweep_config(ref_app_config: RefAppConfig) -> a121.SubsweepConfig:
    step_length = 2  # Typically not very important, we just want a few points close to the sensor.

    start_point = int(round(ref_app_config.obstruction_start_m / APPROX_BASE_STEP_LENGTH_M))
    num_points = max(
        int(
            round(
                np.ceil(
                    (ref_app_config.obstruction_end_m - ref_app_config.obstruction_start_m)
                    / (step_length * APPROX_BASE_STEP_LENGTH_M)
                )
            )
        ),
        1,
    )
    conf_obstruction = a121.SubsweepConfig(
        start_point=start_point,
        num_points=num_points,
        step_length=step_length,
        profile=a121.Profile.PROFILE_1,
        hwaas=16,
    )
    return conf_obstruction


def get_sensor_configs(ref_app_config: RefAppConfig) -> Dict[str, a121.SensorConfig]:
    """
    Returns a dict with sensor configs for the main functionality as well as for obstruction (if applicable)
    Even if the configs will be subsweeps, the function returns separate SensorConfig Objects.
    The intended usage is to be able to easily lift the obstruction functionality.
    In addition, some functions (get_distances) require the SensorConfig to be supplied.
    """

    start_point = int(round(ref_app_config.range_start_m / APPROX_BASE_STEP_LENGTH_M))
    profile = a121.Profile(ref_app_config.profile)

    fwhm = ENVELOPE_FWHM_M[profile]

    step_length_ideal = int(round((fwhm / APPROX_BASE_STEP_LENGTH_M)))
    if step_length_ideal >= 24:
        step_length = int(np.floor(step_length_ideal / 24) * 24)
    else:
        step_length = step_length_ideal
        while step_length > 2 and 24 % step_length != 0:
            step_length -= 1
    hwaas = ref_app_config.hwaas
    num_points = max(
        int(
            round(
                np.ceil(
                    (ref_app_config.range_end_m - ref_app_config.range_start_m)
                    / (step_length * APPROX_BASE_STEP_LENGTH_M)
                )
            )
        ),
        1,
    )

    subsweeps = []
    conf_base = a121.SubsweepConfig(
        start_point=start_point,
        num_points=num_points,
        step_length=step_length,
        profile=profile,
        enable_tx=True,
        hwaas=hwaas,
    )

    ret = {}
    ret["base_config"] = a121.SensorConfig(subsweeps=[conf_base])

    subsweeps.append(conf_base)

    if ref_app_config.obstruction_detection:
        conf_obstruction = get_obstruction_subsweep_config(ref_app_config)
        ret["obstruction_config"] = a121.SensorConfig(
            subsweeps=[conf_obstruction], sweeps_per_frame=1
        )
        subsweeps.append(conf_obstruction)

    conf = a121.SensorConfig(subsweeps=subsweeps, sweeps_per_frame=1)
    ret["full_config"] = conf

    return ret



# Copyright (c) Acconeer AB, 2023
# All rights reserved

import numpy as np

import acconeer.exptool as et
from acconeer.exptool.a111.algo import presence_detection_sparse


def get_sensor_config():
    config = et.a111.SparseServiceConfig()
    config.profile = et.a111.SparseServiceConfig.Profile.PROFILE_3
    config.sampling_mode = et.a111.SparseServiceConfig.SamplingMode.B
    config.range_interval = [0.18, 1.5]
    config.update_rate = 40
    config.sweeps_per_frame = 32
    config.hw_accelerated_average_samples = 60
    return config


class ProcessingConfiguration(et.configbase.ProcessingConfig):
    VERSION = 1

    slow_motion_threshold = et.configbase.FloatParameter(
        label="Slow motion threshold",
        default_value=2,
        limits=(0, 5),
        updateable=True,
        order=0,
        help=('Level at which the slow motion detection is considered as "Slow motion".'),
    )

    slow_motion_hf_cutoff = et.configbase.FloatParameter(
        label="Slow motion high cutoff freq.",
        unit="Hz",
        default_value=20.0,
        limits=(1, 100),
        logscale=True,
        updateable=True,
        order=10,
        help=(
            "Cutoff frequency of the low pass filter for the fast filtered sweep mean."
            " No filtering is applied if the cutoff is set over half the frame rate"
            " (Nyquist limit)."
        ),
    )

    slow_motion_lf_cutoff = et.configbase.FloatParameter(
        label="Slow motion low cutoff freq.",
        unit="Hz",
        default_value=0.2,
        limits=(0.01, 1),
        logscale=True,
        updateable=True,
        order=20,
        help="Cutoff frequency of the low pass filter for the slow filtered sweep mean.",
    )

    slow_motion_deviation_time_const = et.configbase.FloatParameter(
        label="Slow motion deviation time const.",
        unit="s",
        default_value=3,
        limits=(0.01, 30),
        logscale=True,
        updateable=True,
        order=30,
        help=(
            "Time constant of the low pass filter for the (slow-motion) deviation between"
            " fast and slow."
        ),
    )

    fast_motion_threshold = et.configbase.FloatParameter(
        label="Fast motion threshold",
        default_value=1.4,
        limits=(0, 2),
        updateable=True,
        order=40,
        help=('Level at which the fast motion detection is considered as "Fast motion".'),
    )

    fast_motion_time_const = et.configbase.FloatParameter(
        label="Fast motion time const.",
        unit="s",
        default_value=0.5,
        limits=(0, 0.5),
        updateable=True,
        order=50,
        help="Time constant for the fast motion part.",
    )

    fast_motion_outlier = et.configbase.BoolParameter(
        label="Fast motion outlier detection",
        default_value=False,
        updateable=True,
        order=60,
        help=(
            "Fast motion detection includes outlier detection to optimize detection in"
            " (i.e.) meeting room."
        ),
    )

    num_removed_pc = et.configbase.IntParameter(
        label="PCA based noise reduction",
        default_value=0,
        limits=(0, 2),
        updateable=False,
        order=70,
        help=(
            "Sets the number of principal components removed in the PCA based noise reduction."
            " Filters out static reflections."
            " Setting to 0 (default) disables the PCA based noise reduction completely."
        ),
    )

    show_data = et.configbase.BoolParameter(
        label="Show data scatter plot",
        default_value=True,
        updateable=True,
        order=100,
        help=(
            "Show the plot of the current data frame along with the fast and slow filtered"
            " mean sweep (used in the slow-motion part)."
        ),
    )

    show_slow = et.configbase.BoolParameter(
        label="Show slow motion",
        default_value=True,
        updateable=True,
        order=110,
        help="Show the slow motion plot.",
        category=et.configbase.Category.ADVANCED,
    )

    show_fast = et.configbase.BoolParameter(
        label="Show fast motion",
        default_value=True,
        updateable=True,
        order=115,
        help="Show the fast motion plot.",
        category=et.configbase.Category.ADVANCED,
    )

    adaptive_threshold = et.configbase.BoolParameter(
        label="Adaptive threshold",
        default_value=True,
        updateable=False,
        order=120,
        help="Changes the threshold array based on the slow and fast motion processing",
        category=et.configbase.Category.ADVANCED,
    )

    show_sectors = et.configbase.BoolParameter(
        label="Show distance sectors",
        default_value=False,
        updateable=True,
        order=130,
    )

    fast_guard_s = et.configbase.FloatParameter(
        label="Fast motion guard interval [s]",
        default_value=160,
        limits=(30, 2000),
        updateable=False,
        order=140,
        help=(
            "Recording time/period in seconds where there should be no fast"
            " motion in the recording period. It records the maximum slow motion threshold"
            " as a new threshold."
        ),
        category=et.configbase.Category.ADVANCED,
    )

    history_length_s = et.configbase.FloatParameter(
        label="History length",
        unit="s",
        default_value=5,
        limits=(1, 20),
        decimals=0,
        order=200,
        category=et.configbase.Category.ADVANCED,
    )

    def check_sensor_config(self, conf):
        alerts = {
            "processing": [],
            "sensor": [],
        }
        if conf.update_rate is None:
            alerts["sensor"].append(et.configbase.Error("update_rate", "Must be set"))

        if not conf.sweeps_per_frame > 3:
            alerts["sensor"].append(et.configbase.Error("sweeps_per_frame", "Must be > 3"))

        return alerts


class Processor:
    def __init__(self, sensor_config, processing_config, session_info, calibration=None):
        # Config of the presence detector
        self.sweeps_per_frame = sensor_config.sweeps_per_frame
        self.depths = et.a111.get_range_depths(sensor_config, session_info)
        self.num_depths = self.depths.size
        self.f = sensor_config.update_rate
        self.fast_weight = 0.5
        self.slow_weight = 1.0 - self.fast_weight
        self.presence_config = presence_detection_sparse.ProcessingConfiguration()
        self.presence_config.intra_frame_weight = self.fast_weight
        self.presence_config.num_removed_pc = processing_config.num_removed_pc
        self.presence_detector = presence_detection_sparse.Processor(
            sensor_config, self.presence_config, session_info
        )
        self.presence_distance_index = 0
        self.presence_distance = 0
        self.information = "no_motion"
        self.presence_score_depth = np.zeros(self.num_depths)
        self.human_window_end_index = 0
        self.fls_pos_window_s = 3
        self.fls_neg_window_s = 0.6
        self.Qfactor = 3
        self.percentage_quar1 = 25
        self.percentage_quar3 = 75
        self.InterQuar_no_human = 0.15
        self.cal_state = True
        self.fast_outlier_detected = False
        self.start_cal_index = 0
        self.fast_guard_s = processing_config.fast_guard_s
        self.adaptive_threshold = processing_config.adaptive_threshold
        self.fast_outliers_window_s = 20
        self.fast_outliers = np.linspace(
            1,
            processing_config.fast_motion_threshold,
            num=int(np.ceil(self.fast_outliers_window_s * self.f)),
        )
        self.fls_pos_frame = np.zeros(int(np.ceil(self.fls_pos_window_s * self.f)))
        self.fls_neg_frame = np.zeros(int(np.ceil(self.fls_neg_window_s * self.f)))
        self.slow_history = np.zeros(int(round(self.f * processing_config.history_length_s)))
        self.fast_history = np.zeros(int(round(self.f * processing_config.history_length_s)))
        self.update_processing_config(processing_config)

    def update_processing_config(self, processing_config):
        self.human_window_s = 5 * processing_config.slow_motion_deviation_time_const
        self.cal_window_s = self.fast_guard_s - self.human_window_s
        self.end_cal_index = self.start_cal_index + self.cal_window_s * self.f
        self.fast_motion_threshold = processing_config.fast_motion_threshold
        self.slow_motion_threshold = processing_config.slow_motion_threshold
        self.threshold_stored = np.ones(self.num_depths) * self.slow_motion_threshold
        self.threshold_array = np.ones(self.num_depths) * self.slow_motion_threshold
        self.fast_motion_outlier = processing_config.fast_motion_outlier
        self.presence_config.inter_frame_fast_cutoff = processing_config.slow_motion_hf_cutoff
        self.presence_config.inter_frame_slow_cutoff = processing_config.slow_motion_lf_cutoff
        self.presence_config.inter_frame_deviation_time_const = (
            processing_config.slow_motion_deviation_time_const
        )
        self.presence_config.intra_frame_time_const = processing_config.fast_motion_time_const
        self.presence_detector.update_processing_config(self.presence_config)

    def process(self, data, data_info):
        presence_result = self.presence_detector.process(data, data_info)
        fast_motion = presence_result["intra"] / self.fast_weight
        slow_motion = presence_result["inter"] / self.slow_weight

        # Fast Motion Detector
        max_fast = np.max(fast_motion)
        self.fast_outliers = np.roll(self.fast_outliers, -1)
        self.fast_outliers[-1] = np.minimum(max_fast, self.fast_motion_threshold)
        if self.fast_motion_outlier:
            # Detect fast motion using outlier, based on lower boundary and upper boundary
            Quar1 = np.percentile(self.fast_outliers, self.percentage_quar1)
            Quar3 = np.percentile(self.fast_outliers, self.percentage_quar3)
            InterQuar = (
                self.InterQuar_no_human
                if Quar3 - Quar1 >= self.InterQuar_no_human
                else Quar3 - Quar1
            )
            Lower_boundary = Quar1 - self.Qfactor * InterQuar
            Upper_boundary = Quar3 + self.Qfactor * InterQuar
            self.fast_outlier_detected = max_fast <= Lower_boundary or max_fast >= Upper_boundary
        else:
            self.fast_outlier_detected = False
        fast_motion_detected = self.fast_outlier_detected or max_fast > self.fast_motion_threshold
        fast_presence_distance_index = int(np.argmax(fast_motion))
        fast_presence_distance = self.depths[fast_presence_distance_index]

        # Slow Motion Detector
        slow_motion_detected = any(slow_motion > self.threshold_array)
        slow_presence_distance_index = int(np.argmax(slow_motion - self.threshold_array))
        slow_presence_distance = self.depths[slow_presence_distance_index]

        # False Positive False Negative estimation
        self.fls_pos_frame = np.roll(self.fls_pos_frame, -1)
        self.fls_neg_frame = np.roll(self.fls_neg_frame, -1)
        self.fls_pos_frame[-1] = int(not (fast_motion_detected) and slow_motion_detected)
        self.fls_neg_frame[-1] = int(fast_motion_detected and not (slow_motion_detected))
        fls_neg_detected = sum(self.fls_neg_frame) >= len(self.fls_neg_frame)
        fls_pos_detected = (
            sum(self.fls_pos_frame) >= len(self.fls_pos_frame)
            and self.presence_detector.update_index >= self.human_window_end_index
        )

        if fast_motion_detected:
            self.human_window_end_index = (
                self.human_window_s * self.f + self.presence_detector.update_index
            )
            if self.cal_state:
                self.start_cal_index = self.human_window_end_index

            if slow_motion_detected:
                self.fls_pos_frame = np.zeros(len(self.fls_pos_frame))
                self.fls_neg_frame = np.zeros(len(self.fls_neg_frame))

        if fls_neg_detected:
            self.threshold_array = np.ones(self.num_depths) * self.slow_motion_threshold

        if fls_pos_detected:
            if not (self.cal_state):
                self.start_cal_index = self.presence_detector.update_index
            self.cal_state = True

        # Calibration Steps
        if self.presence_detector.update_index == self.end_cal_index:
            self.cal_state = False
            self.fls_pos_frame = np.zeros(len(self.fls_pos_frame))
            self.fls_neg_frame = np.zeros(len(self.fls_neg_frame))
        elif (
            self.presence_detector.update_index > self.start_cal_index
            and self.presence_detector.update_index < self.end_cal_index
        ):
            self.threshold_stored = np.maximum(self.threshold_stored, slow_motion)
        elif self.presence_detector.update_index == self.start_cal_index:
            self.threshold_stored = np.ones(self.num_depths) * self.slow_motion_threshold
        self.end_cal_index = self.start_cal_index + self.cal_window_s * self.f

        # Option to implement the adaptive value threshold
        if self.adaptive_threshold and (self.presence_detector.update_index >= self.end_cal_index):
            self.threshold_array = self.threshold_stored

        if fast_motion_detected:  # Fast motion is prioritized due to the responsiveness
            presence_detected = True
            self.presence_distance_index = fast_presence_distance_index
            self.presence_distance = fast_presence_distance
            self.information = "fast_motion"
        elif slow_motion_detected:
            presence_detected = True
            self.presence_distance_index = slow_presence_distance_index
            self.presence_distance = slow_presence_distance
            self.information = "slow_motion"
        else:
            presence_detected = False
            self.presence_distance = 0
            self.information = "no_motion"

        if (self.end_cal_index - self.presence_detector.update_index) / self.f <= 10 and (
            self.end_cal_index - self.presence_detector.update_index
        ) / self.f > 0:
            self.information = "adapting_threshold"

        # Histories
        self.fast_history = np.roll(self.fast_history, -1)
        self.fast_history[-1] = fast_motion_detected
        self.slow_history = np.roll(self.slow_history, -1)
        self.slow_history[-1] = slow_motion_detected

        out_data = {
            "frame": presence_result["frame"],
            "fast": presence_result["fast"],
            "slow": presence_result["slow"],
            "fast_motion": fast_motion,
            "slow_motion": slow_motion,
            "presence_distance_index": self.presence_distance_index,
            "presence_distance": self.presence_distance,
            "information": self.information,
            "presence_detected": presence_detected,
            "fls_neg_s": sum(self.fls_neg_frame) / self.f,
            "fls_pos_s": sum(self.fls_pos_frame) / self.f,
            "fast_history": self.fast_history,
            "slow_history": self.slow_history,
            "threshold_array": self.threshold_array,
        }

        return out_data



# Copyright (c) Acconeer AB, 2022-2025
# All rights reserved

from __future__ import annotations

import logging
from typing import Optional

from PySide6.QtGui import QTextCursor
from PySide6.QtWidgets import (
    QDialog,
    QGridLayout,
    QPushButton,
    QTextEdit,
    QWidget,
)


_WRONG_CREDENTIALS_MSG = "<font color='red'>Incorrect username (email) or password</font>"

log = logging.getLogger(__name__)


class UserMessageDialog(QDialog):
    def __init__(
        self,
        title: str,
        message: Optional[str],
        confirmation: str,
        parent: Optional[QWidget] = None,
    ) -> None:
        super().__init__(parent)

        self.setMinimumWidth(450)
        self.setMinimumHeight(200)

        layout = QGridLayout(self)

        self.message_text = QTextEdit(self)
        self.message_text.setReadOnly(True)

        self.ok_button = QPushButton(confirmation, self)
        self.ok_button.setEnabled(True)
        self.ok_button.clicked.connect(self._on_ok)

        # fmt: off
        # Grid layout:                      row, col, rspan, cspan
        layout.addWidget(self.message_text, 0,   0,   4,     12)    # noqa: E241
        layout.addWidget(self.ok_button,    4,   0,   1,     2)     # noqa: E241
        # fmt: on

        self.setLayout(layout)

        self.setWindowTitle(title)
        if message is not None:
            self.set_message(message)

    def set_message(self, message: str) -> None:
        self.message_text.clear()
        self.message_text.insertHtml(message)
        self.message_text.moveCursor(QTextCursor.MoveOperation.Start)

    def _on_ok(self) -> None:
        self.accept()



# Copyright (c) Acconeer AB, 2024-2025
# All rights reserved

from __future__ import annotations

import collections
from typing import Optional

import attrs
import h5py
import numpy as np
import numpy.typing as npt
from attributes_doc import attributes_doc

from acconeer.exptool import a121
from acconeer.exptool.a121.algo import (
    APPROX_BASE_STEP_LENGTH_M,
    AlgoProcessorConfigBase,
    ProcessorBase,
    get_distances_m,
)


@attributes_doc
@attrs.mutable(kw_only=True)
class ProcessorConfig(AlgoProcessorConfigBase):
    _MIN_START_POINT = 32  # Set to avoid direct leakage using profile 1
    _MIN_MEASURABLE_DISTANCE_M = _MIN_START_POINT * APPROX_BASE_STEP_LENGTH_M

    bin_start_m: float = attrs.field(default=0.15)
    """Minimum detection distance from sensor."""
    bin_end_m: float = attrs.field(default=1.00)
    """Maximum detection distance from sensor."""
    threshold: float = attrs.field(default=0.3)
    """Threshold for which the standard deviation of the phase should be below."""
    distance_sequence_n: int = attrs.field(default=4)
    """Number of distance needed in sequence below threshold to be presumed as waste level."""
    median_filter_length: int = attrs.field(default=5)
    """Length of the median filter used to stabilize the level result."""

    def _collect_validation_results(
        self, config: a121.SessionConfig
    ) -> list[a121.ValidationResult]:
        validation_results: list[a121.ValidationResult] = []

        if not self.bin_start_m < self.bin_end_m:
            validation_results.append(
                a121.ValidationError(
                    self,
                    "bin_start_m",
                    "Minimum detection distance must be shorter than maximum detection distance.",
                )
            )

        if config.sensor_config.sweeps_per_frame < 4:
            validation_results.append(
                a121.ValidationError(
                    config.sensor_config,
                    "sweeps_per_frame",
                    "Sweeps per frame must be > 3.",
                )
            )

        min_measured_point = min(ss.start_point for ss in config.sensor_config.subsweeps)
        max_measured_point = max(
            ss.start_point + (ss.num_points - 1) * ss.step_length
            for ss in config.sensor_config.subsweeps
        )
        min_measured_distance_m = min_measured_point * APPROX_BASE_STEP_LENGTH_M
        max_measured_distance_m = max_measured_point * APPROX_BASE_STEP_LENGTH_M
        measured_range_str = f"{min_measured_distance_m:.2f}-{max_measured_distance_m:.2f}m"

        for subsweep in config.sensor_config.subsweeps:
            if subsweep.start_point < self._MIN_START_POINT:
                validation_results.append(
                    a121.ValidationError(
                        subsweep,
                        "start_point",
                        f"Start point must be greater or equal to {self._MIN_START_POINT}",
                    )
                )

        if self.bin_start_m < self._MIN_MEASURABLE_DISTANCE_M:
            validation_results.append(
                a121.ValidationError(
                    self,
                    "bin_start_m",
                    f"Bin start must be greater or equal to {self._MIN_MEASURABLE_DISTANCE_M} m",
                )
            )

        if not min_measured_distance_m <= self.bin_start_m <= max_measured_distance_m:
            bin_msg = f"Bin start is outside measured range ({measured_range_str})"
            validation_results += [a121.ValidationError(self, "bin_start_m", bin_msg)]

        if not min_measured_distance_m <= self.bin_end_m <= max_measured_distance_m:
            bin_msg = f"Bin end is outside measured range ({measured_range_str})"
            validation_results += [a121.ValidationError(self, "bin_end_m", bin_msg)]

        for (prev_num, prev), (curr_num, curr) in zip(
            enumerate(config.sensor_config.subsweeps, start=1),
            enumerate(config.sensor_config.subsweeps[1:], start=2),
        ):
            if not prev.start_point < curr.start_point:
                msg = (
                    "Start point needs to be larger than the "
                    + f"previous subsweep's start point (at least {prev.start_point + 1})"
                )
                validation_results += [a121.ValidationError(curr, "start_point", msg)]

            prev_end_point = prev.start_point + prev.step_length * (prev.num_points - 1)
            if not prev_end_point < curr.start_point:
                msg = f"Subsweeps {prev_num} and {curr_num} overlap."
                prev_msg = f"{msg} Adjust end point."
                curr_msg = f"{msg} Adjust start point."
                validation_results += [
                    a121.ValidationError(prev, "num_points", prev_msg),
                    a121.ValidationError(prev, "step_length", prev_msg),
                    a121.ValidationError(curr, "start_point", curr_msg),
                ]

        return validation_results


@attrs.frozen(kw_only=True)
class ProcessorContext: ...


@attrs.frozen(kw_only=True)
class ProcessorExtraResult:
    """
    Contains information for visualization in ET.
    """

    phase_std: npt.NDArray[np.float64]
    """The standard deviation of the phase from which the distance is detemined."""
    distance_m: Optional[float]
    """Distance from sensor to fill level."""


@attrs.frozen(kw_only=True)
class ProcessorResult:
    level_m: Optional[float]
    """Fill level in the bin in meters."""
    level_percent: Optional[int]
    """Fill level in the bin in percentage."""
    extra_result: ProcessorExtraResult
    """Extra result used for visualization."""


class Processor(ProcessorBase[ProcessorResult]):
    def __init__(
        self,
        *,
        sensor_config: a121.SensorConfig,
        metadata: a121.Metadata,
        processor_config: ProcessorConfig,
        subsweep_indexes: Optional[list[int]] = None,
        context: Optional[ProcessorContext] = None,
    ) -> None:
        self.bin_end_m = processor_config.bin_end_m
        self.bin_start_m = processor_config.bin_start_m
        self.median_filter_length = processor_config.median_filter_length
        self.threshold = processor_config.threshold
        self.distance_sequence_n = processor_config.distance_sequence_n

        assert self.bin_start_m < self.bin_end_m

        self.distance_history: collections.deque[float] = collections.deque(
            [], maxlen=self.median_filter_length
        )

        self.distances_m = get_distances_m(sensor_config, metadata)

    def find_distance_m(self, phase_stds: npt.NDArray[np.float64]) -> float:
        stable_phases = phase_stds < self.threshold

        potential_distances = (
            float(self.distances_m[idx])
            for idx in range(len(stable_phases))
            if stable_phases[idx : idx + self.distance_sequence_n].all()
        )

        return next(potential_distances, np.nan)

    def process(self, result: a121.Result) -> ProcessorResult:
        # center the phase distribution around zero
        sweep_sums = np.sum(result.frame, axis=0)
        phases = np.angle(result.frame * np.conj(sweep_sums))

        # calculate standard deviation of the phase at each distance over multiple sweeps
        phase_std = np.std(phases, axis=0)
        distance_m = self.find_distance_m(phase_std)

        self.distance_history.append(distance_m)

        if np.all(np.isnan(list(self.distance_history))):
            level_m = None
            level_percent = None
            filtered_distance = None
        else:
            filtered_distance = float(np.nanmedian(list(self.distance_history)))
            level_m = self.bin_end_m - filtered_distance
            level_percent = int(round(level_m / (self.bin_end_m - self.bin_start_m) * 100, 0))
            if level_percent < 0:
                level_percent = 0

        extra_result = ProcessorExtraResult(phase_std=phase_std, distance_m=filtered_distance)

        return ProcessorResult(
            level_m=level_m,
            level_percent=level_percent,
            extra_result=extra_result,
        )

    def update_config(self, config: ProcessorConfig) -> None: ...


def _load_algo_data(algo_group: h5py.Group) -> ProcessorConfig:
    processor_config = ProcessorConfig.from_json(algo_group["processor_config"][()])
    return processor_config



# Copyright (c) Acconeer AB, 2022
# All rights reserved

import numpy as np
from matplotlib.colors import LinearSegmentedColormap

from PySide6 import QtGui

import pyqtgraph as pg

import acconeer.exptool as et


class PGUpdater:
    def __init__(self, sensor_config, processing_config, session_info):
        self.sensor_config = sensor_config
        self.processing_config = processing_config

        self.depths = et.a111.get_range_depths(sensor_config, session_info)
        self.depth_res = session_info["step_length_m"]
        self.xs = np.tile(self.depths, sensor_config.sweeps_per_frame)
        self.smooth_limits = et.utils.SmoothLimits(sensor_config.update_rate)

    def setup(self, win):
        win.setWindowTitle("Acconeer sparse example")

        # For history images:
        rate = self.sensor_config.update_rate
        xlabel = "Frames" if rate is None else "Time (s)"
        x_scale = 1.0 if rate is None else 1.0 / rate
        y_scale = self.depth_res
        x_offset = -self.processing_config.history_length * x_scale
        y_offset = self.depths[0] - 0.5 * self.depth_res

        self.data_plots = []
        self.scatters = []
        self.data_history_ims = []
        self.presence_history_ims = []

        for i in range(len(self.sensor_config.sensor)):
            data_plot = win.addPlot(title="Sparse data", row=0, col=i)
            data_plot.setMenuEnabled(False)
            data_plot.setMouseEnabled(x=False, y=False)
            data_plot.hideButtons()
            data_plot.showGrid(x=True, y=True)
            data_plot.setLabel("bottom", "Depth (m)")
            data_plot.setLabel("left", "Amplitude")
            scatter = pg.ScatterPlotItem(size=10)
            data_plot.addItem(scatter)

            cmap_cols = ["steelblue", "lightblue", "#f0f0f0", "moccasin", "darkorange"]
            cmap = LinearSegmentedColormap.from_list("mycmap", cmap_cols)
            cmap._init()
            lut = (cmap._lut * 255).view(np.ndarray).astype(np.uint8)

            self.data_history_plot = win.addPlot(title="Data history", row=1, col=i)
            self.data_history_plot.setMenuEnabled(False)
            self.data_history_plot.setMouseEnabled(x=False, y=False)
            self.data_history_plot.hideButtons()
            data_history_im = pg.ImageItem(autoDownsample=True)
            data_history_im.setLookupTable(lut)
            self.data_history_plot.addItem(data_history_im)
            self.data_history_plot.setLabel("bottom", xlabel)
            self.data_history_plot.setLabel("left", "Depth (m)")

            self.presence_history_plot = win.addPlot(title="Movement history", row=2, col=i)
            self.presence_history_plot.setMenuEnabled(False)
            self.presence_history_plot.setMouseEnabled(x=False, y=False)
            self.presence_history_plot.hideButtons()
            presence_history_im = pg.ImageItem(autoDownsample=True)
            presence_history_im.setLookupTable(et.utils.pg_mpl_cmap("viridis"))
            self.presence_history_plot.addItem(presence_history_im)
            self.presence_history_plot.setLabel("bottom", xlabel)
            self.presence_history_plot.setLabel("left", "Depth (m)")

            for im in [presence_history_im, data_history_im]:
                im.resetTransform()
                tr = QtGui.QTransform()
                tr.translate(x_offset, y_offset)
                tr.scale(x_scale, y_scale)
                im.setTransform(tr)

            self.data_plots.append(data_plot)
            self.scatters.append(scatter)
            self.data_history_ims.append(data_history_im)
            self.presence_history_ims.append(presence_history_im)

        self.setup_is_done = True
        self.update_processing_config()

    def update_processing_config(self, processing_config=None):
        if processing_config is None:
            processing_config = self.processing_config
        else:
            self.processing_config = processing_config

        if not self.setup_is_done:
            return

        self.presence_history_plot.setVisible(processing_config.show_move_history_plot)
        self.data_history_plot.setVisible(processing_config.show_data_history_plot)

    def update(self, d):
        data_limits = self.smooth_limits.update(d["data"])

        for i in range(len(self.sensor_config.sensor)):
            ys = d["data"][i].flatten()
            self.scatters[i].setData(self.xs, ys)
            self.data_plots[i].setYRange(*data_limits)

            data_history_adj = d["data_history"][:, i] - 2**15
            sign = np.sign(data_history_adj)
            data_history_adj = np.abs(data_history_adj)
            data_history_adj /= data_history_adj.max()
            data_history_adj = np.power(data_history_adj, 1 / 2.2)  # gamma correction
            data_history_adj *= sign
            self.data_history_ims[i].updateImage(data_history_adj, levels=(-1.05, 1.05))

            m = np.max(d["presence_history"][:, i]) * 1.1
            self.presence_history_ims[i].updateImage(d["presence_history"][:, i], levels=(0, m))



# Copyright (c) Acconeer AB, 2022-2025
# All rights reserved

from __future__ import annotations

import abc
import logging
import typing as t

from PySide6.QtWidgets import QVBoxLayout

import pyqtgraph as pg

from acconeer.exptool.app.new.app_model import AppModel
from acconeer.exptool.app.new.backend import GeneralMessage

from .ui_plugin_base import UiPluginBase


log = logging.getLogger(__name__)


class PlotPluginBase(UiPluginBase):
    """
    A basic widget with the following signals connected:

    AppModel signal -> method
    --------------------------------------
    sig_notify -> on_app_model_update (no-op hook)
    sig_load_plugin -> on_load_plugin
    sig_message_plot_plugin -> handle_message
    sig_backend_state_changed -> on_backend_state_update (no-op hook)
    """

    def __init__(self, app_model: AppModel) -> None:
        super().__init__(app_model)

        self.app_model = app_model

        self.app_model.sig_message_plot_plugin.connect(self.handle_message)
        self.app_model.sig_notify.connect(self.on_app_model_update)
        self.app_model.sig_backend_state_changed.connect(self.on_backend_state_update)
        self.app_model.sig_load_plugin.connect(self.on_load_plugin)

    def on_load_plugin(self, plugin_spec: t.Optional[t.Any]) -> None:
        if plugin_spec is None:
            self.app_model.sig_message_plot_plugin.disconnect(self.handle_message)
            self.app_model.sig_notify.disconnect(self.on_app_model_update)
            self.app_model.sig_backend_state_changed.disconnect(self.on_backend_state_update)
            self.app_model.sig_load_plugin.disconnect(self.on_load_plugin)

    def on_app_model_update(self, app_model: AppModel) -> None:
        pass

    def on_backend_state_update(self, state: t.Optional[t.Any]) -> None:
        pass

    @abc.abstractmethod
    def handle_message(self, message: GeneralMessage) -> None:
        """Handles messages with recipient=="plot_plugin" from the BackendPlugin"""
        pass

    @abc.abstractmethod
    def draw(self) -> None:
        """Gets called with a set frequency (see PluginPlotArea)"""
        pass


# def handle_message(self, message: GeneralMessage) -> None:
#     if message.kwargs is None:
#         raise RuntimeError("Plot message needs non-None kwargs")

#     if message.name == "setup":
#         self.setup(**message.kwargs)
#         self._is_setup = True
#     elif message.name == "plot":
#         self._plot_job = message.kwargs
#     else:
#         log.warn(f"{self.__class__.__name__} got an unsupported command: {message.name!r}.")

# def draw(self) -> None:
#     if not self._is_setup or self._plot_job is None:
#         return

#     try:
#         self.draw_plot_job(**self._plot_job)
#     finally:
#         self._plot_job = None


class PgPlotPlugin(PlotPluginBase):
    """
    Expands PlotPluginBase by adding a pyqtgraph GraphicsLayout
    to the widget. The GraphicsLayout is accessible via the
    plot_layout attribute
    """

    def __init__(self, app_model: AppModel) -> None:
        super().__init__(app_model=app_model)

        self.plot_widget = pg.GraphicsLayoutWidget()
        self.plot_layout = self.plot_widget.ci

        layout = QVBoxLayout()

        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self.plot_widget)

        self.setLayout(layout)



# Copyright (c) Acconeer AB, 2022-2025
# All rights reserved

SDK_VERSION = "1.12.0"

# Make these visible under the a121 package to not break api
from acconeer.exptool._cli import ExampleArgumentParser, get_client_args
from acconeer.exptool._core.communication.client import ClientError, ServerError
from acconeer.exptool._core.entities import (
    ClientInfo,
    Criticality,
    MockInfo,
    SerialInfo,
    SocketInfo,
    USBInfo,
    ValidationError,
    ValidationResult,
    ValidationWarning,
)
from acconeer.exptool._core.int_16_complex import (
    complex_array_to_int16_complex,
    int16_complex_array_to_complex,
)

from ._core import (
    _H5PY_STR_DTYPE,
    PRF,
    Client,
    H5Record,
    H5Recorder,
    IdleState,
    InMemoryRecord,
    Metadata,
    PersistentRecord,
    Profile,
    Record,
    Recorder,
    RecordError,
    Result,
    SensorCalibration,
    SensorConfig,
    SensorInfo,
    ServerInfo,
    SessionConfig,
    StackedResults,
    SubsweepConfig,
    iterate_extended_structure,
    iterate_extended_structure_values,
    load_record,
    open_record,
    save_record,
    save_record_to_h5,
    zip3_extended_structures,
    zip_extended_structures,
)
from ._core_ext import ReplaySessionsExhaustedError, _ReplayingClient, _StopReplay
from ._perf_calc import (
    _SensorPerformanceCalc,
    _SessionPerformanceCalc,
    get_point_overhead_duration,
    get_sample_duration,
)



# Copyright (c) Acconeer AB, 2022-2025
# All rights reserved

from __future__ import annotations

import json
import typing as t
import warnings

import attrs
from attributes_doc import attributes_doc

from acconeer.exptool import opser
from acconeer.exptool._core.class_creation.formatting import pretty_dict_line_strs
from acconeer.exptool._core.entities.validation_result import (
    ValidationError,
    ValidationResult,
    ValidationWarning,
)
from acconeer.exptool.a121._core.utils import (
    EntityJSONEncoder,
    convert_value,
    is_divisor_of,
    is_multiple_of,
)

from .config_enums import PRF, Profile


SPARSE_IQ_PPC = 24

# The `converter` argument to `attrs.field` influences the signature of `__init__`
# - (https://www.attrs.org/en/stable/api.html#converters)
# This is mirrored in LSPs and the docs.
#
# `convert_value` is a generic function which needs to be specialiced to `int`s for use here
# Plain `Profile` / `PRF` functions as converters in runtime, but no type is in
# the signature of `__init__`.


def int_converter(value: int) -> int:
    return convert_value(value, factory=int)


def profile_converter(profile: Profile) -> Profile:
    return Profile(profile)


def prf_converter(prf: PRF) -> PRF:
    return PRF(prf)


_T = t.TypeVar("_T")


def _copy_docstring_from(thing: t.Any) -> t.Callable[[_T], _T]:
    """Modifies the decorated function by copying over the __doc__ of `thing`"""

    def inner(decoratee: _T) -> _T:
        decoratee.__doc__ = thing.__doc__
        return decoratee

    return inner


@attributes_doc
@attrs.mutable(kw_only=True)
class SubsweepConfig:
    """Subsweep configuration

    The subsweep config represents a 1-1 mapping to the RSS service subsweep config.

    Normally used as a part of the :attr:`SensorConfig`. Multiple group configurations are required
    when certain parameters cannot be configured in subsweep config.
    """

    MAX_HWAAS = 511

    start_point: int = attrs.field(default=80, converter=int_converter)
    """Starting point of the sweep

    The starting point of the sweep. The corresponding start in millimeter is approximately
    ``start_point`` * 2.5 mm.
    """

    num_points: int = attrs.field(
        default=160,
        converter=int_converter,
        validator=[attrs.validators.ge(1)],
    )
    """Number of data points to measure

    The number of data points to measure in a sweep.
    """

    step_length: int = attrs.field(default=1, converter=int_converter)
    """Step length in a sweep

    This sets the number of steps to have between each data point.

    The corresponding distance between each data point is ``step_length`` * 2.5 mm.
    """

    profile: Profile = attrs.field(default=Profile.PROFILE_3, converter=profile_converter)
    """Profile

    Each profile consists of a number of settings for the sensor that configures the RX and TX
    paths. Lower profiles have higher depth resolution while higher profiles have higher radar
    loop gain.
    """

    hwaas: int = attrs.field(
        default=8,
        converter=int_converter,
        validator=[attrs.validators.ge(1), attrs.validators.le(511)],
    )
    """Hardware accelerated average samples (HWAAS)

    Each data point can be sampled several times and the sensor hardware then produces an
    average value of those samples. The time needed to measure a sweep is roughly proportional
    to the number of averaged samples. Hence, if there is a need to obtain a higher update
    rate, HWAAS could be decreased but this leads to lower SNR.

    HWAAS must be between 1 and 511 inclusive.
    """

    receiver_gain: int = attrs.field(
        default=16,
        converter=int_converter,
        validator=[attrs.validators.ge(0), attrs.validators.le(23)],
    )
    """Receiver gain setting

    Must be a value between 0 and 23 inclusive where 23 is the highest gain and 0 the lowest.

    Lower gain gives higher SNR. However, too low gain may result in quantization, lowering
    SNR. Too high gain may result in saturation, corrupting the data.
    """

    enable_tx: bool = attrs.field(default=True, converter=bool)
    """Enable or disable the transmitter

    If set to True, TX is enabled. This will enable the radio transmitter. By turning the
    transmitter off the RX noise floor can be measured.
    """

    enable_loopback: bool = attrs.field(default=False, converter=bool)
    """Enable or disable loopback

    Enabling loopback will activate an internal route between TX and RX in the sensor.
    The signal will take this route instead of being transmitted out of the sensor.

    Note, loopback can't be enabled together with profile 2.
    """

    phase_enhancement: bool = attrs.field(default=False, converter=bool)
    """Enable or disable phase enhancement

    If enabled, the data phase will be enhanced such that coherent distance filtering can be
    applied. Given a single reflection from an object, the phase will appear as "flat" around
    the amplitude peak.

    Enabling the phase enhancement increases the processing execution time.
    """

    iq_imbalance_compensation: bool = attrs.field(default=False, converter=bool)
    """Enable or disable IQ imbalance compensation

    If enabled, reduces undesirable amplitude variations over distance.

    Enabling IQ imbalance compensation increases the processing execution time.
    """

    _prf: PRF = attrs.field(
        default=PRF.PRF_15_6_MHz,
        converter=prf_converter,
    )

    @property
    @_copy_docstring_from(PRF)
    def prf(self) -> PRF:
        return self._prf

    @prf.setter
    def prf(self, value: PRF) -> None:
        self._prf = PRF(value)

    def _collect_validation_results(self) -> list[ValidationResult]:
        APPROX_BASE_STEP_LENGTH = 2.5e-3
        validation_results: list[ValidationResult] = []

        if self.enable_loopback and self.profile == Profile.PROFILE_2:
            validation_results.extend(
                [
                    ValidationError(
                        self, "enable_loopback", "Enable loopback is incompatible with Profile 2."
                    ),
                    ValidationError(
                        self, "profile", "Enable loopback is incompatible with Profile 2."
                    ),
                ]
            )

        if self.prf == PRF.PRF_19_5_MHz:
            FORBIDDEN_PROFILES = [
                Profile.PROFILE_2,
                Profile.PROFILE_3,
                Profile.PROFILE_4,
                Profile.PROFILE_5,
            ]

            if self.profile in FORBIDDEN_PROFILES:
                validation_results.extend(
                    [
                        ValidationError(
                            self, "prf", "19.5 MHz PRF is only compatible with profile 1."
                        ),
                        ValidationError(
                            self,
                            "profile",
                            f"Profile {self.profile.value} is not supported with 19.5 MHz PRF.",
                        ),
                    ]
                )

        end_point_m = (
            self.start_point + ((self.num_points - 1) * self.step_length)
        ) * APPROX_BASE_STEP_LENGTH
        if end_point_m > self.prf.maximum_measurable_distance:
            validation_results.extend(
                [
                    ValidationError(
                        self,
                        "prf",
                        f"PRF is too high for the measuring end point ({end_point_m:.3f}m), "
                        + "try lowering the PRF.",
                    ),
                    ValidationError(
                        self,
                        "num_points",
                        f"Measuring range is too long for PRF (max {self.prf.mmd:.2f}m), "
                        + "try decreasing the number of points.",
                    ),
                    ValidationError(
                        self,
                        "step_length",
                        f"Measuring range is too long for PRF (max {self.prf.mmd:.2f}m), "
                        + "try decreasing the step length.",
                    ),
                ]
            )

        if self.start_point < -275:
            validation_results.append(
                ValidationError(
                    self,
                    "start_point",
                    "Start point must be larger than or equal to -275",
                )
            )

        return validation_results

    def validate(self) -> None:
        """Performs self-validation

        :raises ValidationError: If anything is invalid.
        """
        for validation_result in self._collect_validation_results():
            try:
                raise validation_result
            except ValidationWarning as vw:
                warnings.warn(vw.message)

    @staticmethod
    @step_length.validator
    def step_length_validator(instance: t.Any, _: t.Any, step_length: int) -> None:
        if not (
            is_divisor_of(SPARSE_IQ_PPC, step_length) or is_multiple_of(SPARSE_IQ_PPC, step_length)
        ):
            msg = f"Step length must be a divisor or multiple of {SPARSE_IQ_PPC}"
            raise ValueError(msg)

    def to_dict(self) -> dict[str, t.Any]:
        return {k.strip("_"): v for k, v in attrs.asdict(self).items()}

    @classmethod
    def from_dict(cls, d: dict[str, t.Any]) -> SubsweepConfig:
        return SubsweepConfig(**d)

    def to_json(self) -> str:
        return json.dumps(self.to_dict(), cls=EntityJSONEncoder)

    @classmethod
    def from_json(cls, json_str: str) -> SubsweepConfig:
        return cls.from_dict(json.loads(json_str))

    def _pretty_str_lines(self, index: t.Optional[int] = None) -> list[str]:
        lines = []
        index_str = "" if index is None else f" @ index {index}"
        lines.append(f"{type(self).__name__}{index_str}:")
        lines.extend(pretty_dict_line_strs(self.to_dict()))
        return lines

    def __str__(self) -> str:
        return "\n".join(self._pretty_str_lines())


opser.register_json_presentable(SubsweepConfig)



# Copyright (c) Acconeer AB, 2022-2025
# All rights reserved
from __future__ import annotations

import argparse
import ctypes
import functools
import logging
import re
import sys
import time
import typing as t
from importlib.resources import as_file, files

import qdarktheme

from PySide6 import QtCore, QtGui
from PySide6.QtWidgets import QApplication

import pyqtgraph as pg

from acconeer.exptool.app import resources
from acconeer.exptool.utils import config_logging

from ._enums import ConnectionState, PluginGeneration, PluginState
from .app_model import AppModel
from .backend import Backend, GenBackend, MpBackend
from .plugin_loader import import_and_register_plugin_module, load_plugins
from .storage import remove_config_dir, remove_temp_dir
from .ui.app_model_viewer import AppModelViewer
from .ui.main_window import MainWindow


_LOG = logging.getLogger(__name__)


def main() -> None:
    parser = _ExptoolArgumentParser()
    args = parser.parse_args()
    config_logging(args)

    if args.tasks:
        any_autoconnect = (
            args.usb_or_serial_autoconnect
            or args.simulated_autoconnect
            or args.socket_autoconnect is not None
        )
        if not any_autoconnect:
            parser.print_usage()
            print("ERROR: One of the 'autoconnect' options are required when using '--tasks'")
            sys.exit(1)

        if args.plugin_key is None:
            parser.print_usage()
            print("ERROR: The '--plugin-key' option is required when using '--tasks'")
            sys.exit(1)

    if args.purge_cache:
        remove_config_dir()
        remove_temp_dir()
        print("Cache purged")
        sys.exit(0)

    if sys.platform == "win32" or sys.platform == "cygwin":
        ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID("acconeer.exptool")

    if QApplication.instance() is None:
        QApplication.setHighDpiScaleFactorRoundingPolicy(
            QtCore.Qt.HighDpiScaleFactorRoundingPolicy.PassThrough,
        )

    # The cast necessitated of a miss in the typing stubs.
    app: QApplication = t.cast(QApplication, QApplication.instance()) or QApplication(sys.argv)

    # Make sure to wrap urls in single quotation marks to avoid errors parsing stylesheet
    stylesheet = qdarktheme.load_stylesheet("light")
    search_pattern = r"url\((.*?)\)"
    replace_pattern = r"url('\1')"
    stylesheet = re.sub(search_pattern, replace_pattern, stylesheet)

    app.setStyleSheet(stylesheet)
    app.setAttribute(QtCore.Qt.ApplicationAttribute.AA_UseHighDpiPixmaps)

    if args.plugin_modules is not None:
        for plugin_module_name in args.plugin_modules:
            import_and_register_plugin_module(plugin_module_name)

    plugins = load_plugins()
    plugin_keys = [p.key for p in plugins]

    if args.plugin_key is not None and args.plugin_key not in plugin_keys:
        parser.print_usage()
        print(f"ERROR: Could not find plugin with key {args.plugin_key!r}")
        print(f"ERROR: Available plugin keys: {plugin_keys}")
        sys.exit(1)

    backend: Backend
    if args.same_process:
        backend = GenBackend()
    else:
        backend = MpBackend()
    backend.start()

    model = AppModel(
        backend,
        plugins,
        args.usb_or_serial_autoconnect,
        args.socket_autoconnect,
        args.simulated_autoconnect,
    )
    model.start()

    app.aboutToQuit.connect(model.stop)
    app.aboutToQuit.connect(backend.stop)

    pg.setConfigOption("background", "w")
    pg.setConfigOption("foreground", "k")
    pg.setConfigOption("leftButtonPan", False)
    pg.setConfigOptions(antialias=True)

    with as_file(files(resources) / "icon.png") as path:
        app.setWindowIcon(_pixmap_to_icon(QtGui.QPixmap(str(path))))

    mw = MainWindow(model)
    mw.show()

    if args.amv:
        app_model_viewer = AppModelViewer(model)
        app_model_viewer.show()
        mw.sig_closing.connect(app_model_viewer.close)

    specified_generation = PluginGeneration[args.generation_str.upper()]
    model.set_plugin_generation(specified_generation)

    if args.plugin_key is not None:
        found_plugin = model.find_plugin(args.plugin_key, specified_generation)

        if found_plugin is None:
            _LOG.warning(f"Could not find plugin with key {args.plugin_key!r}")
        else:
            if args.tasks:
                tasks = tuple(str(t) for t in args.tasks)

                model.sig_notify.connect(
                    functools.partial(
                        _send_startup_tasks, tasks=tasks, deadline=time.monotonic() + 5
                    ),
                    QtCore.Qt.ConnectionType.SingleShotConnection,
                )

            model.load_plugin(found_plugin)

    model.broadcast()

    sys.exit(app.exec())


def _send_startup_tasks(app_model: AppModel, tasks: tuple[str, ...], deadline: float) -> None:
    """Uses tail-end recursion to offer passed 'tasks' once the app
    is in the correct state
    """
    if not tasks:
        _LOG.info("No more tasks to run.")
        return

    seconds_left = deadline - time.monotonic()
    if seconds_left < 0:
        _LOG.warning(f"Startup tasks timed out. Giving up on tasks {list(tasks)}.")
        return

    (task_to_send, *rest_of_tasks) = tasks

    def _on_error(exception: Exception, traceback_format_exc: t.Optional[str] = None) -> None:
        _LOG.warning(
            f"Startup tasks that won't execute due to the failure of {task_to_send!r}: {rest_of_tasks}"
        )
        app_model.emit_error(exception, traceback_format_exc)

    def _retry() -> None:
        app_model.sig_notify.connect(
            functools.partial(_send_startup_tasks, tasks=tasks, deadline=deadline),
            QtCore.Qt.ConnectionType.SingleShotConnection,
        )

    retry_str = f"Retrying ... ({seconds_left:.2f}s left)"

    if app_model.connection_state is not ConnectionState.CONNECTED:
        _LOG.info(f"Cannot execute {task_to_send!r}: No sensor connected yet. {retry_str}")
        _retry()
    elif app_model.plugin_state.is_busy:
        _LOG.info(f"Cannot execute {task_to_send!r}: Plugin is busy. {retry_str}")
        _retry()
    elif app_model.plugin_state is not PluginState.LOADED_IDLE:
        _LOG.info(f"Cannot execute {task_to_send!r}: No plugin loaded yet. {retry_str}")
        _retry()
    else:
        _LOG.info(f"Executing {task_to_send!r} ...")
        app_model.put_task(
            (task_to_send, {}),
            on_ok=lambda: _send_startup_tasks(app_model, tuple(rest_of_tasks), deadline),
            on_error=_on_error,
        )


class _ExptoolArgumentParser(argparse.ArgumentParser):
    def __init__(self) -> None:
        super().__init__(prog="python -m acconeer.exptool.app.new")

        self.add_argument(
            "--purge-cache",
            "--purge-config",
            action="store_true",
            help="Remove cache files.",
        )
        self.add_argument(
            "--portable",
            action="store_true",
            help=argparse.SUPPRESS,  # makes option hidden
        )

        plugin_group = self.add_argument_group(title="plugins")
        plugin_group.add_argument(
            "--plugin-module",
            dest="plugin_modules",
            metavar="module",
            action="append",  # "append" => --plugin-module X --plugin-module Y => [X, Y]
            help=(
                "Allows you to load an arbitrary plugin given a python module "
                + "(installed or in your working directory) NOTE! Accepted argument "
                + "is not a path (e.g. not 'my_processor/latest/plugin.py'), it's a "
                + "python module (e.g. 'my_processor.latest.plugin'). "
                + "This option can be repeated."
            ),
        )

        startup_group = self.add_argument_group(title="startup")
        connect_group = startup_group.add_mutually_exclusive_group()
        connect_group.add_argument(
            "--autoconnect",
            dest="usb_or_serial_autoconnect",
            action="store_true",
            help="Enable the auto connect feature for USB or serial devices.",
        )
        connect_group.add_argument(
            "--socket-autoconnect",
            metavar="ip",
            help="Connect to 'ip' on startup.",
        )
        connect_group.add_argument(
            "--simulated-autoconnect",
            action="store_true",
            help="Connect to the sensor simulation at startup.",
        )
        startup_group.add_argument(
            "--plugin-key",
            dest="plugin_key",
            metavar="key",
            type=str,
            help="Load the plugin with given key on startup. E.g. 'sparse_iq' or 'distance_detector'.",
        )
        startup_group.add_argument(
            "--generation",
            dest="generation_str",
            choices=[PluginGeneration.A121.value],
            default="a121",
            type=str,
            help="Generation that is selected on startup. Also used in conjunction with '--plugin-key' to find plugins.",
        )
        startup_group.add_argument(
            "--tasks",
            nargs="+",
            default=[],
            help=(
                "A chain of zero-argument backend tasks ('@is_task') to execute at startup, "
                + "e.g. 'start_session' or 'calibrate_detector start_session'. "
                + "If you run into issues, turn on verbose logs ('-v') for more information."
            ),
        )

        dbg_group = self.add_argument_group(title="debugging")
        dbg_group.add_argument(
            "--same-process",
            action="store_true",
            help="Run the application backend in the main thread. Deteriorates performance but enables backend plugin debugging (e.g. with breakpoint())",
        )
        dbg_group.add_argument(
            "--amv",
            action="store_true",
            help="Start the AppModelViewer alongside Exploration Tool",
        )

        verbosity_group = self.add_mutually_exclusive_group(required=False)
        verbosity_group.add_argument(
            "-v",
            "--verbose",
            action="store_true",
        )
        verbosity_group.add_argument(
            "-vv",
            "--debug",
            action="store_true",
        )
        verbosity_group.add_argument(
            "-q",
            "--quiet",
            action="store_true",
        )


def _pixmap_to_icon(pixmap: QtGui.QPixmap) -> QtGui.QIcon:
    size = max(pixmap.size().height(), pixmap.size().width())

    square_pixmap = QtGui.QPixmap(size, size)
    square_pixmap.fill(QtGui.QRgba64.fromRgba(0, 0, 0, 0))

    painter = QtGui.QPainter(square_pixmap)
    painter.drawPixmap(
        (square_pixmap.size().width() - pixmap.size().width()) // 2,
        (square_pixmap.size().height() - pixmap.size().height()) // 2,
        pixmap,
    )
    painter.end()

    scaled_pixmap = square_pixmap.scaled(
        256,
        256,
        aspectMode=QtCore.Qt.AspectRatioMode.KeepAspectRatio,
        mode=QtGui.Qt.TransformationMode.SmoothTransformation,
    )

    return QtGui.QIcon(scaled_pixmap)



# Copyright (c) Acconeer AB, 2022-2024
# All rights reserved

from __future__ import annotations

import logging
from typing import NoReturn, Optional, Tuple, Type, TypeVar, Union

import attrs
import typing_extensions as te

from acconeer.exptool._core.communication import (
    BufferedLink,
    ClientCreationError,
    ClientError,
    ExploreSerialLink,
    Message,
    MessageStream,
)
from acconeer.exptool._core.communication.client import ServerError
from acconeer.exptool._core.communication.communication_protocol import messages
from acconeer.exptool._core.communication.communication_protocol.messages.log_message import (
    ServerLog,
)
from acconeer.exptool._core.communication.links.helpers import ensure_connected_link
from acconeer.exptool._core.communication.unwrap_ticks import unwrap_ticks
from acconeer.exptool._core.entities import ClientInfo
from acconeer.exptool.a121._core.entities import (
    Metadata,
    Result,
    SensorCalibration,
    SensorConfig,
    ServerInfo,
    SessionConfig,
)
from acconeer.exptool.a121._core.utils import (
    create_extended_structure,
    iterate_extended_structure,
    unextend,
)
from acconeer.exptool.a121._perf_calc import _SessionPerformanceCalc

from .client import Client
from .exploration_protocol import (
    ExplorationProtocol,
    get_exploration_protocol,
)
from .exploration_protocol import (
    messages as a121_messages,
)
from .utils import get_calibrations_provided


_MessageT = TypeVar("_MessageT", bound=Message)
log = logging.getLogger(__name__)


class ExplorationClient(Client, register=True):
    _link: BufferedLink
    _protocol: Type[ExplorationProtocol]
    _tick_unwrapper: TickUnwrapper
    _server_info: Optional[ServerInfo]
    _result_queue: list[list[dict[int, Result]]]
    _log_queue: list[ServerLog]

    @classmethod
    def open(
        cls,
        ip_address: Optional[str] = None,
        tcp_port: Optional[int] = None,
        serial_port: Optional[str] = None,
        usb_device: Optional[Union[str, bool]] = None,
        mock: Optional[bool] = None,
        override_baudrate: Optional[int] = None,
        flow_control: bool = True,
        generation: Optional[str] = "a121",
    ) -> te.Self:
        if generation != "a121":
            raise ClientCreationError

        if mock is not None:
            raise ClientCreationError

        client_info = ClientInfo._from_open(
            ip_address=ip_address,
            tcp_port=tcp_port,
            override_baudrate=override_baudrate,
            serial_port=serial_port,
            usb_device=usb_device,
            flow_control=flow_control,
        )

        return cls(client_info=client_info)

    def __init__(
        self,
        client_info: ClientInfo = ClientInfo(mock=None),
        _override_protocol: Optional[Type[ExplorationProtocol]] = None,
    ) -> None:
        super().__init__(client_info)
        self._tick_unwrapper = TickUnwrapper()
        self._server_info = None
        self._log_queue = []
        self._closed = False
        self._crashing = False

        self._protocol = ExplorationProtocol

        (self._link, self._client_info) = ensure_connected_link(self.client_info)
        self._server_stream = MessageStream(
            self._link,
            self._protocol,
            message_handler=self._handle_messages,
            link_error_callback=self._close_before_reraise,
        )

        self._server_info = self._retrieve_server_info()

        if self._server_info.connected_sensors == []:
            self._link.disconnect()
            msg = "Exploration server is running but no sensors are detected."
            raise ClientError(msg)

        if _override_protocol is None:
            most_suitable_protocol = get_exploration_protocol(self.server_info.parsed_rss_version)

            self._server_stream.protocol = most_suitable_protocol
            self._protocol = most_suitable_protocol
        else:
            self._server_stream.protocol = _override_protocol
            self._protocol = _override_protocol

        self._update_baudrate()

    def _close_before_reraise(self, exception: Exception) -> NoReturn:
        self._crashing = True
        self.close()
        raise exception

    def _retrieve_server_info(self) -> ServerInfo:
        self._server_stream.send_command(self._protocol.get_system_info_command())
        system_info_response = self._server_stream.wait_for_message(messages.SystemInfoResponse)

        sensor = system_info_response.system_info.get("sensor")
        if sensor != "a121":
            self.close()
            msg = f"Wrong sensor version, expected a121 but got {sensor}"
            raise ClientError(msg)

        self._server_stream.send_command(self._protocol.get_sensor_info_command())
        sensor_info_response = self._server_stream.wait_for_message(
            a121_messages.SensorInfoResponse
        )

        return ServerInfo(
            rss_version=system_info_response.system_info["rss_version"],
            sensor_count=system_info_response.system_info["sensor_count"],
            ticks_per_second=system_info_response.system_info["ticks_per_second"],
            hardware_name=system_info_response.system_info.get("hw", None),
            sensor_infos=sensor_info_response.sensor_infos,
            max_baudrate=system_info_response.system_info.get("max_baudrate"),
        )

    def _update_baudrate(self) -> None:
        # Only Change baudrate for ExploreSerialLink
        if not isinstance(self._link, ExploreSerialLink):
            return

        if self.client_info.serial is None:
            return

        DEFAULT_BAUDRATE = 115200
        overridden_baudrate = self.client_info.serial.override_baudrate
        max_baudrate = self.server_info.max_baudrate
        baudrate_to_use = self.server_info.max_baudrate or DEFAULT_BAUDRATE

        # Override baudrate?
        if overridden_baudrate is not None and max_baudrate is not None:
            # Valid Baudrate?
            if overridden_baudrate > max_baudrate:
                msg = f"Cannot set a baudrate higher than {max_baudrate}"
                raise ClientError(msg)
            elif overridden_baudrate < DEFAULT_BAUDRATE:
                msg = f"Cannot set a baudrate lower than {DEFAULT_BAUDRATE}"
                raise ClientError(msg)
            baudrate_to_use = overridden_baudrate

        # Do not change baudrate if DEFAULT_BAUDRATE
        if baudrate_to_use == DEFAULT_BAUDRATE:
            return

        self._server_stream.send_command(self._protocol.set_baudrate_command(baudrate_to_use))
        _ = self._server_stream.wait_for_message(messages.SetBaudrateResponse)

        self._link.baudrate = baudrate_to_use

    def setup_session(  # type: ignore[override]
        self,
        config: Union[SensorConfig, SessionConfig],
        calibrations: Optional[dict[int, SensorCalibration]] = None,
    ) -> Union[Metadata, list[dict[int, Metadata]]]:
        if self.session_is_started:
            msg = "Session is currently running, can't setup."
            raise ClientError(msg)

        if isinstance(config, SensorConfig):
            config = SessionConfig(config)

        config.validate()

        self._calibrations_provided = get_calibrations_provided(config, calibrations)
        self._session_config = config

        self._server_stream.send_command(self._protocol.setup_command(config, calibrations))
        setup_response = self._server_stream.wait_for_message(a121_messages.SetupResponse)

        self._metadata = [
            {
                sensor_id: metadata
                for metadata, sensor_id in zip(metadata_group, config_group.keys())
            }
            for metadata_group, config_group in zip(
                setup_response.grouped_metadatas, self._session_config.groups
            )
        ]
        self._sensor_calibrations = setup_response.sensor_calibrations

        if self.session_config.extended:
            return self._metadata
        else:
            return unextend(self._metadata)

    def _handle_messages(self, message: Message) -> None:
        if type(message) is messages.LogMessage:
            self._log_queue.append(message.message)
        elif type(message) is a121_messages.EmptyResultMessage:
            msg = "Received an empty Result from Server."
            raise RuntimeError(msg)
        elif type(message) is messages.ErroneousMessage:
            last_error = ""
            for log in self._log_queue:
                if log.level == "ERROR" and "exploration_server" not in log.module:
                    last_error = f" ({log.log})"
            msg = f"{message}{last_error}"
            raise ServerError(msg)

    def start_session(self) -> None:
        self._assert_session_setup()

        if self.session_is_started:
            msg = "Session is already started."
            raise ClientError(msg)

        assert self._session_config is not None

        pc = _SessionPerformanceCalc(self._session_config, self._metadata)

        try:
            # Use max of the calculate duration and update/frame rate to guarantee sufficient
            # timeout.
            timeout_duration = max(pc.update_duration, 1 / pc.update_rate)
            # Increase timeout if update rate is very low, otherwise keep default
            timeout = max(1.5 * timeout_duration + 1.0, self._link.DEFAULT_TIMEOUT)
        except Exception:
            timeout = self._link.DEFAULT_TIMEOUT

        self._link.timeout = timeout

        self._server_stream.send_command(self._protocol.start_streaming_command())
        _ = self._server_stream.wait_for_message(messages.StartStreamingResponse)

        self._recorder_start_session()
        self._session_is_started = True

    def get_next(self) -> Union[Result, list[dict[int, Result]]]:  # type: ignore[override]
        self._assert_session_started()

        result_message = self._server_stream.wait_for_message(a121_messages.ResultMessage)

        if self._metadata is None:
            msg = f"{self} has no metadata"
            raise RuntimeError(msg)

        if self._server_info is None:
            msg = f"{self} has no system info"
            raise RuntimeError(msg)

        if self._session_config is None:
            msg = f"{self} has no session config"
            raise RuntimeError(msg)

        extended_results = result_message.get_extended_results(
            tps=self._server_info.ticks_per_second,
            metadata=self._metadata,
            config_groups=self._session_config.groups,
        )

        extended_results = self._tick_unwrapper.unwrap_ticks(extended_results)

        self._recorder_sample(extended_results)
        return self._return_results(extended_results)

    def stop_session(self) -> None:
        self._assert_session_started()

        self._server_stream.send_command(self._protocol.stop_streaming_command())
        _ = self._server_stream.wait_for_message(
            messages.StopStreamingResponse,
            timeout_s=self._link.timeout + 1,
        )

        self._link.timeout = self._link.DEFAULT_TIMEOUT
        self._session_is_started = False
        self._recorder_stop_session()
        self._log_queue.clear()

    def close(self) -> None:
        if self._closed:
            return

        try:
            if self.session_is_started:
                if self._crashing:
                    self._session_is_started = False
                    self._recorder_stop_session()
                else:
                    self.stop_session()
        except Exception:
            raise
        finally:
            self._tick_unwrapper = TickUnwrapper()
            self._server_info = None
            self._metadata = None
            self._log_queue.clear()
            self._link.disconnect()
            self._closed = True

    @property
    def connected(self) -> bool:
        return self._server_info is not None

    @property
    def server_info(self) -> ServerInfo:
        self._assert_connected()

        assert self._server_info is not None  # Should never happend if client is connected
        return self._server_info


class TickUnwrapper:
    """Wraps unwrap_ticks to be applied over extended results"""

    def __init__(self) -> None:
        self.next_minimum_tick: Optional[int] = None

    def unwrap_ticks(self, extended_results: list[dict[int, Result]]) -> list[dict[int, Result]]:
        result_items = list(iterate_extended_structure(extended_results))
        ticks = [result.tick for _, _, result in result_items]
        unwrapped_ticks, self.next_minimum_tick = unwrap_ticks(ticks, self.next_minimum_tick)

        def f(result_item: Tuple[int, int, Result], updated_tick: int) -> Tuple[int, int, Result]:
            group_index, sensor_id, result = result_item
            updated_result = attrs.evolve(result, tick=updated_tick)
            return (group_index, sensor_id, updated_result)

        return create_extended_structure(map(f, result_items, unwrapped_ticks))
